from typing import Optional

import numpy as np

from mygrad.operation_base import _NoValue
from mygrad.tensor_base import Tensor, asarray, implements_numpy_override
from mygrad.typing import ArrayLike

from .ops import Where

__all__ = ["where"]


@implements_numpy_override()
def where(
    condition: ArrayLike,
    x: ArrayLike = _NoValue,
    y: ArrayLike = _NoValue,
    *,
    constant: Optional[bool] = None
) -> Tensor:
    """
    where(condition, [x, y])

    Return elements chosen from `x` or `y` depending on `condition`.

    .. note::
        When only ``condition`` is provided, this function is a shorthand for
        ``np.asarray(condition).nonzero()``. The rest of this
        documentation covers only the case where all three arguments are
        provided.

    This docstring was adapted from that of ``numpy.where``.

    Parameters
    ----------
    condition : ArrayLike, bool
        Where True, yield `x`, otherwise yield ``y``. ``x``, ``y``
        and `condition` need to be broadcastable to some shape.

    x : ArrayLike
        Values from which to chosen where ``condition`` is ``True``.

    y : ArrayLike
       Values from which to chosen where ``condition`` is ``False``.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    out : mygrad.Tensor
        A tensor with elements from `x` where `condition` is True, and elements
        from `y` elsewhere.

    Examples
    --------
    >>> import mygrad as mg
    >>> a = mg.arange(10)
    >>> a
    Tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> mg.where(a < 5, a, 10*a)
    Tensor([ 0,  1,  2,  3,  4, 50, 60, 70, 80, 90])

    This can be used on multidimensional tensors too:

    >>> mg.where([[True, False], [True, True]],
    ...          [[1, 2], [3, 4]],
    ...          [[9, 8], [7, 6]])
    Tensor([[1, 8],
            [3, 4]])

    The shapes of x, y, and the condition are broadcast together:

    >>> x, y = np.ogrid[:3, :4]
    >>> mg.where(x < y, x, 10 + y)  # both x and 10+y are broadcast
    Tensor([[10,  0,  0,  0],
            [10, 11,  1,  1],
            [10, 11, 12,  2]])

    >>> a = mg.Tensor([[0, 1, 2],
    ...                [0, 2, 4],
    ...                [0, 3, 6]])
    >>> mg.where(a < 4, a, -1)  # -1 is broadcast
    Tensor([[ 0,  1,  2],
            [ 0,  2, -1],
            [ 0,  3, -1]])
    """
    if x is _NoValue and y is _NoValue:
        return np.where(asarray(condition))

    if x is _NoValue or y is _NoValue:
        raise ValueError("either both or neither of x and y should be given")

    return Tensor._op(
        Where, x, y, op_kwargs={"condition": condition}, constant=constant
    )



from numbers import Real
from typing import Optional

import numpy as np

import mygrad._utils.graph_tracking as _tracking
from mygrad.nnet.activations import softmax
from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor, asarray
from mygrad.typing import ArrayLike

from ._utils import check_loss_inputs

__all__ = ["softmax_focal_loss", "focal_loss"]


class FocalLoss(Operation):
    r"""Returns the per-datum focal loss as described in https://arxiv.org/abs/1708.02002
    which is given by -ɑ(1-p)ˠlog(p).

    Extended Description
    --------------------
    The focal loss is given by

    .. math::
        \frac{1}{N}\sum\limits_{1}^{N}-\alpha \hat{y}_i(1-p_i)^\gamma\log(p_i)

    where :math:`N` is the number of elements in `x` and `y` and :math:`\hat{y}_i` is
    one where :math:`i` is the label of the element :math:`y_i` and 0 elsewhere. That is,
    if the label :math:`y_k` is 1 and there are four possible label values, then
    :math:`\hat{y}_k = (0, 1, 0, 0)`.

    It is recommended in the paper that you normalize by the number of foreground samples.
    """

    def __call__(self, class_probs, targets, alpha, gamma):
        """
        Parameters
        ----------
        class_probs : mygrad.Tensor, shape=(N, C)
            The C class scores for each of the N pieces of data.

        targets : Union[mygrad.Tensor, ArrayLike], shape=(N,)
            The correct class indices, in [0, C), for each datum.

        alpha : Real
            The ɑ weighting factor in the loss formulation.

        gamma : Real
            The ɣ focusing parameter.

        Returns
        -------
        numpy.ndarray
            The per-datum focal loss.
        """
        if isinstance(targets, Tensor):  # pragma: nocover
            targets = targets.data

        check_loss_inputs(class_probs, targets)

        self.variables = (class_probs,)
        self.label_locs = (range(len(class_probs)), targets)

        class_probs = asarray(class_probs)
        pc = class_probs[self.label_locs]
        one_m_pc = np.clip(1 - pc, a_min=0, a_max=1)
        log_pc = np.log(pc)

        one_m_pc_gamma = one_m_pc ** gamma
        loss = -(alpha * one_m_pc_gamma * log_pc)

        if not _tracking.TRACK_GRAPH:
            return loss

        self.back = np.zeros(class_probs.shape, dtype=np.float64)

        if np.isclose(gamma, 0, atol=1e-15):
            self.back[self.label_locs] -= alpha / pc
            return loss

        # dL/dp = -alpha * ( (1 - p)**g / p - g * (1 - p)**(g - 1) * log(p) )
        #
        # term 1: (1 - p)**g / p
        term1 = one_m_pc_gamma / pc  # (1 - p)**g / p

        # term 2: - g * (1 - p)**(g - 1) * log(p)
        if np.isclose(gamma, 1, rtol=1e-15):
            term2 = -log_pc
        elif gamma < 1:
            # For g < 1 and p -> 1, the 2nd term -> 0 via L'Hôpital's rule
            term2 = np.zeros(pc.shape, dtype=np.float64)
            pc_not_1 = ~np.isclose(one_m_pc, 0, atol=1e-15)
            term2[pc_not_1] = (
                -gamma * one_m_pc[pc_not_1] ** (gamma - 1) * log_pc[pc_not_1]
            )
        else:
            term2 = -gamma * one_m_pc ** (gamma - 1) * log_pc

        self.back[self.label_locs] -= alpha * (term1 + term2)
        return loss

    def backward_var(self, grad, index, **kwargs):
        self.back[self.label_locs] *= grad
        return self.back


def focal_loss(
    class_probs: ArrayLike,
    targets: ArrayLike,
    *,
    alpha: float = 1,
    gamma: float = 0,
    constant: Optional[bool] = None,
) -> Tensor:
    r"""Return the per-datum focal loss.

    Parameters
    ----------
    class_probs : ArrayLike, shape=(N, C)
        The C class probabilities for each of the N pieces of data.
        Each value is expected to lie on (0, 1]

    targets : ArrayLike, shape=(N,)
        The correct class indices, in [0, C), for each datum.

    alpha : Real, optional (default=1)
        The ɑ weighting factor in the loss formulation.

    gamma : Real, optional (default=0)
        The ɣ focusing parameter. Note that for Ɣ=0 and ɑ=1, this is cross-entropy loss.
        Must be a non-negative value.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor, shape=(N,)
        The per-datum focal loss.

    Notes
    -----
    The formulation for the focal loss introduced in https://arxiv.org/abs/1708.02002.
    It is given by -ɑ(1-p)ˠlog(p).


    The focal loss for datum-:math:`i` is given by

    .. math::
        -\alpha \hat{y}_i(1-p_i)^\gamma\log(p_i)

    where :math:`\hat{y}_i` is one in correspondence to the label associated with the
    datum and 0 elsewhere. That is, if the label :math:`y_k` is 2 and
    there are four possible label values, then :math:`\hat{y}_k = (0, 0, 1, 0)`.

    It is recommended in the paper that you normalize by the number of foreground samples.
    """
    if not isinstance(gamma, Real) or gamma < 0:
        raise ValueError(f"`gamma` must be a non-negative number, got: {gamma}")

    return Tensor._op(
        FocalLoss, class_probs, op_args=(targets, alpha, gamma), constant=constant
    )


def softmax_focal_loss(
    scores: ArrayLike,
    targets: ArrayLike,
    *,
    alpha: float = 1,
    gamma: float = 0,
    constant: Optional[bool] = None,
) -> Tensor:
    r"""
    Applies the softmax normalization to the input scores before computing the
    per-datum focal loss.

    Parameters
    ----------
    scores : ArrayLike, shape=(N, C)
        The C class scores for each of the N pieces of data.

    targets : ArrayLike, shape=(N,)
        The correct class indices, in [0, C), for each datum.

    alpha : Real, optional (default=1)
        The ɑ weighting factor in the loss formulation.

    gamma : Real, optional (default=0)
        The ɣ focusing parameter. Note that for Ɣ=0 and ɑ=1, this is cross-entropy loss.
        Must be a non-negative value.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor, shape=(N,)
        The per-datum focal loss.

    Notes
    -----
    The formulation for the focal loss introduced in https://arxiv.org/abs/1708.02002.
    It is given by -ɑ(1-p)ˠlog(p).

    The focal loss for datum-:math:`i` is given by

    .. math::
        -\alpha \hat{y}_i(1-p_i)^\gamma\log(p_i)

    where :math:`\hat{y}_i` is one in correspondence to the label associated with the
    datum and 0 elsewhere. That is, if the label :math:`y_k` is 2 and
    there are four possible label values, then :math:`\hat{y}_k = (0, 0, 1, 0)`.

    It is recommended in the paper that you normalize by the number of foreground samples.
    """
    return focal_loss(
        softmax(scores), targets=targets, alpha=alpha, gamma=gamma, constant=constant
    )



from typing import Optional

import numpy as np

from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike


class Sigmoid(Operation):
    def __call__(self, a):
        self.variables = (a,)
        x = np.asarray(-1.0 * a.data)
        np.exp(x, out=x)
        x += 1
        np.reciprocal(x, out=x)
        self.sigmoid = x
        return self.sigmoid

    def backward_var(self, grad, index, **kwargs):
        return grad * self.sigmoid * (1.0 - self.sigmoid)


def sigmoid(x: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """Applies the sigmoid activation function::

      f(x) = 1 / (1 + exp(-x))

    Parameters
    ----------
    x : ArrayLike
        sigmoid is applied element-wise on ``x``.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import sigmoid
    >>> x = mg.linspace(-5, 5, 10)
    >>> sigmoid(x)
    Tensor([0.00669285, 0.02005754, 0.0585369 , 0.1588691 , 0.36457644,
        0.63542356, 0.8411309 , 0.9414631 , 0.97994246, 0.99330715])

    .. plot::

       >>> import mygrad as mg
       >>> from mygrad.nnet.activations import sigmoid
       >>> import matplotlib.pyplot as plt
       >>> x = mg.linspace(-10, 10, 100)
       >>> y = sigmoid(x)
       >>> plt.title("sigmoid(x)")
       >>> y.backward()
       >>> plt.plot(x, x.grad, label="df/dx")
       >>> plt.plot(x, y, label="f(x)")
       >>> plt.legend()
       >>> plt.grid()
       >>> plt.show()"""
    return Tensor._op(Sigmoid, x, constant=constant)



import numpy as np

from mygrad.operation_base import BinaryUfunc, UnaryUfunc

__all__ = [
    "Exp",
    "Exp2",
    "Expm1",
    "Log",
    "Log2",
    "Log10",
    "Log1p",
    "Logaddexp",
    "Logaddexp2",
]


class Exp(UnaryUfunc):
    numpy_ufunc = np.exp

    def backward_var(self, grad, index, **kwargs):
        return grad * np.exp(self.variables[index].data)


class Exp2(UnaryUfunc):
    numpy_ufunc = np.exp2

    def backward_var(self, grad, index, **kwargs):
        return grad * np.exp2(self.variables[index].data) * np.log(2)


class Expm1(UnaryUfunc):
    numpy_ufunc = np.expm1

    def backward_var(self, grad, index, **kwargs):
        return grad * np.exp(self.variables[index].data)


class Logaddexp(BinaryUfunc):
    numpy_ufunc = np.logaddexp

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:
            return grad / (1 + np.exp(b.data - a.data))
        elif index == 1:
            return grad / (1 + np.exp(a.data - b.data))
        else:  # pragma: no cover
            raise IndexError(f"Back-propagation through tensor-{index}")


class Logaddexp2(BinaryUfunc):
    numpy_ufunc = np.logaddexp2

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:
            return grad / (1 + 2 ** (b.data - a.data))
        elif index == 1:
            return grad / (1 + 2 ** (a.data - b.data))
        else:  # pragma: no cover
            raise IndexError(f"Back-propagation through tensor-{index}")


class Log(UnaryUfunc):
    numpy_ufunc = np.log

    def backward_var(self, grad, index, **kwargs):
        return grad / self.variables[index].data


class Log2(UnaryUfunc):
    numpy_ufunc = np.log2

    def backward_var(self, grad, index, **kwargs):
        return grad / (self.variables[index].data * np.log(2))


class Log10(UnaryUfunc):
    numpy_ufunc = np.log10

    def backward_var(self, grad, index, **kwargs):
        return grad / (self.variables[index].data * np.log(10))


class Log1p(UnaryUfunc):
    numpy_ufunc = np.log1p

    def backward_var(self, grad, index, **kwargs):
        return grad / (1 + self.variables[index].data)



from typing import Optional, Tuple, Union

import numpy as np

from mygrad.math._special import logsumexp as _logsumexp
from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike


def _softmax(x, kwargs):

    if x.ndim > 0 and x.size > 0:
        x = x - x.max(**kwargs)
        target = x.astype(float) if issubclass(x.dtype.type, np.integer) else x

        target = np.exp(x, out=target)
        target /= target.sum(**kwargs)
    else:
        target = x.astype(float) if issubclass(x.dtype.type, np.integer) else x
        target = np.ones_like(target)
    return target


class Softmax(Operation):
    def __call__(self, a, axis=-1):
        self.variables = (a,)
        x = a.data

        self._kw = dict(axis=axis, keepdims=True)
        self._cached_output = _softmax(x, self._kw)
        return self._cached_output

    def backward_var(self, grad, index, **kwargs):
        _ = self.variables[index]  # check index error
        soft = self._cached_output
        sg = soft * grad
        return sg - soft * np.sum(sg, **self._kw)


def softmax(
    x: ArrayLike,
    axis: Union[None, int, Tuple[int, ...]] = -1,
    *,
    constant: Optional[bool] = None
) -> Tensor:
    r"""
    Applies the softmax activation function::

        f(x) = exp(x) / sum( exp(x) )

    Computes the softmax over one or more axes of an ND-tensor.

    Parameters
    ----------
    x : array_like

    axis : Union[None, int, Tuple[int, ...]], optional (default=-1)
        The axis/axes over which to compute the softmax.
        By default, the softmax is computed over the trailing axis.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Notes
    -----
    - :math:`N` is the number of samples in the batch.
    - :math:`C` is the number of possible classes for which scores are provided.

    This implements a numerically-stable version of softmax, however
    log-softmax is still the more numerically stable activation function.

    Given the shape-:math:`(N, C)` tensor of scores, ``x``, the softmax classification
    probabilities are computed. That is, the score for class-:math:`k` of a given datum
    (:math:`s_{k}`) is normalized using the 'softmax' transformation:

    .. math::
        p_{k} = \frac{e^{s_k}}{\sum_{i=1}^{C}{e^{s_i}}}

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import softmax
    >>> x = mg.Tensor([[ 2.,  2.,  2.],
    ...                [2E50, 2E50,  1E50]])
    >>> softmax(x)
    Tensor([[0.33333333, 0.33333333, 0.33333333],
            [0.5       , 0.5       , 0.        ]])
    """
    return Tensor._op(Softmax, x, op_kwargs=dict(axis=axis), constant=constant)


class LogSoftmax(Operation):
    scalar_only = True

    def __call__(self, a, axis=-1):
        self.variables = (a,)
        x = a.data

        self._kw = dict(axis=axis, keepdims=True)
        return x - _logsumexp(x, **self._kw)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        x = a.data
        soft = _softmax(x, self._kw)
        return grad - soft * np.sum(grad, **self._kw)


def logsoftmax(
    x: ArrayLike,
    axis: Union[None, int, Tuple[int, ...]] = -1,
    *,
    constant: Optional[bool] = None
) -> Tensor:
    r"""
    Applies the log-softmax activation function::

        f(x) = log ( exp(x) / sum( exp(x) ) )

    Computes the log-softmax over one or more axes of an ND-tensor.

    Parameters
    ----------
    x : ArrayLike

    axis : Union[None, int, Tuple[int, ...]], optional (default=-1)
        The axis/axes over which to compute the log-softmax.
        By default, the log-softmax is computed over the trailing axis.

    constant : constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    log_softmax : mygrad.Tensor
        Tensor with same shape as ``x``

    Notes
    -----
    - :math:`N` is the number of samples in the batch.
    - :math:`C` is the number of possible classes for which scores are provided.

    This implements a numerically-stable version of log-softmax, compared
    to the naive implementation using ``mygrad.log``, ``mygrad.exp``, and
    ``mygrad.sum``.

    Given the shape-:math:`(N, C)` tensor of scores, ``x``, the softmax classification
    probabilities are computed. That is, the score for class-:math:`k` of a given datum
    (:math:`s_{k}`) is normalized using the 'softmax' transformation:

    .. math::
        p_{k} = \log{\frac{e^{s_k}}{\sum_{i=1}^{C}{e^{s_i}}}}

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import logsoftmax
    >>> x = mg.Tensor([[  2.,   2.,    2.],
    ...                [2E50, 2E50,  1E50]])
    >>> logsoftmax(x)
    Tensor([[-1.09861229e+00, -1.09861229e+00, -1.09861229e+00],
            [ 0.00000000e+00,  0.00000000e+00, -1.00000000e+50]])
    """
    return Tensor._op(LogSoftmax, x, op_kwargs=dict(axis=axis), constant=constant)



from itertools import accumulate
from typing import TYPE_CHECKING, Optional

import numpy as np

from mygrad._utils.graph_tracking import TRACK_GRAPH
from mygrad.operation_base import Operation

if TYPE_CHECKING:  # pragma: no cover
    from mygrad import Tensor

__all__ = ["Concatenate"]


class Concatenate(Operation):
    def __call__(
        self,
        *input_vars: "Tensor",
        axis: Optional[int] = 0,
        out: Optional[np.ndarray] = None,
        dtype=None,
    ) -> np.ndarray:
        kwargs = {"out": out, "axis": axis}
        if dtype is not None:
            # compatibility shim: min numpy 1.20
            kwargs["dtype"] = dtype

        out = np.concatenate(tuple(var.data for var in input_vars), **kwargs)

        self.variables = tuple(input_vars)
        if TRACK_GRAPH:
            self.axis = axis
            if axis is not None:
                # need to make sure axis is non-negative so that
                # axis checking during backprop is simplified
                self.axis = axis % out.ndim
                self.indices = list(
                    accumulate(var.data.shape[axis] for var in input_vars)
                )
                self.indices.insert(0, 0)
            else:
                # inputs were flattened
                self.indices = np.cumsum((0,) + tuple(var.size for var in input_vars))
        return out

    def backward_var(self, grad, index, **kwargs) -> np.ndarray:
        var = self.variables[index]
        if self.axis is None:
            return grad[self.indices[index] : self.indices[index + 1]].reshape(
                var.shape
            )

        return grad[
            tuple(
                slice(None, None, None)
                if dim != self.axis
                else slice(self.indices[index], self.indices[index + 1])
                for dim in range(var.data.ndim)
            )
        ]


class Stack(Operation):
    def __call__(
        self,
        *input_vars: "Tensor",
        axis: int = 0,
        out: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        out = np.stack(tuple(var.data for var in input_vars), axis=axis, out=out)

        self.variables = tuple(input_vars)

        if TRACK_GRAPH:
            # need to make sure axis is non-negative so that
            # axis checking during backprop is simplified
            self.axis = axis % out.ndim

        return out

    def backward_var(self, grad, index, **kwargs) -> np.ndarray:
        assert 0 <= index
        var = self.variables[index]

        if var.data.ndim == 0:
            return grad[index]

        return grad[
            tuple(
                slice(None, None, None) if dim != self.axis else index
                for dim in range(var.data.ndim + 1)
            )
        ]



from ._array_like import ArrayLike, Index, Mask, Real
from ._dtype_like import DTypeLike, DTypeLikeReals
from ._shape import Shape

__all__ = [
    "ArrayLike",
    "DTypeLike",
    "DTypeLikeReals",
    "Index",
    "Mask",
    "Real",
    "Shape",
]



from typing import Optional, Union

from numpy import ndarray

from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike, DTypeLikeReals, Mask
from mygrad.ufuncs import ufunc_creator

from .ops import *

__all__ = [
    "sin",
    "sinc",
    "cos",
    "tan",
    "cot",
    "csc",
    "sec",
    "arccos",
    "arccsc",
    "arcsin",
    "arctan",
    "arcsec",
    "arccot",
    "arctan2",
]


@ufunc_creator(Sin)
def sin(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    r"""Trigonometric sine, element-wise.

    This docstring was adapted from that of numpy.sin [1]_

    Parameters
    ----------
    x : ArrayLike
        Angle, in radians (:math:`2 \pi` rad equals 360 degrees).

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    y : Tensor
        The sine of each element of x.

    See Also
    --------
    arcsin, sinh, cos

    Notes
    -----
    The sine is one of the fundamental functions of trigonometry (the
    mathematical study of triangles).  Consider a circle of radius 1
    centered on the origin.  A ray comes in from the :math:`+x` axis, makes
    an angle at the origin (measured counter-clockwise from that axis), and
    departs from the origin.  The :math:`y` coordinate of the outgoing
    ray's intersection with the unit circle is the sine of that angle.  It
    ranges from -1 for :math:`x=3\pi / 2` to +1 for :math:`\pi / 2.`  The
    function has zeroes where the angle is a multiple of :math:`\pi`.
    Sines of angles between :math:`\pi` and :math:`2\pi` are negative.
    The numerous properties of the sine and related functions are included
    in any standard trigonometry text.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.sin.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.sin(mg.pi/2.)
    Tensor(1.0)

    Print sines of an array of angles given in degrees:

    >>> mg.sin(mg.tensor((0., 30., 45., 60., 90.)) * mg.pi / 180. )
    Tensor([ 0.        ,  0.5       ,  0.70710678,  0.8660254 ,  1.        ])
    """
    ...


@ufunc_creator(Cos)
def cos(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Trigonometric cosine, element-wise.

    This docstring was adapted from that of numpy.cos [1]_

    Parameters
    ----------
    x : ArrayLike
        Input array in radians.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    y : Tensor
        The corresponding cosine values.

    Notes
    -----
    If `out` is provided, the function writes the result into it,
    and returns a reference to `out`.  (See Examples)

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.cos.html

    M. Abramowitz and I. A. Stegun, Handbook of Mathematical Functions.
    New York, NY: Dover, 1972.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.cos([0, mg.pi/2, mg.pi])
    Tensor([  1.00000000e+00,   6.12303177e-17,  -1.00000000e+00])
    """
    ...


@ufunc_creator(Tan)
def tan(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Trigonometric tangent, element-wise.

    This docstring was adapted from that of numpy.tan [1]_

    Parameters
    ----------
    x : ArrayLike
        Input array.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    y : Tensor
        The corresponding tangent values.

    Notes
    -----
    If `out` is provided, the function writes the result into it,
    and returns a reference to `out`.  (See Examples)

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.tan.html

    M. Abramowitz and I. A. Stegun, Handbook of Mathematical Functions.
    New York, NY: Dover, 1972.

    Examples
    --------
    >>> import mygrad as mg
    >>> from math import pi
    >>> mg.tan([-pi, pi / 2, pi])
    Tensor([  1.22460635e-16,   1.63317787e+16,  -1.22460635e-16])
    """
    ...


@ufunc_creator(Arcsin)
def arcsin(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Inverse sine, element-wise.

    This docstring was adapted from that of numpy.arcsin [1]_

    Parameters
    ----------
    x : ArrayLike
        `y`-coordinate on the unit circle.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    angle : Tensor
        The inverse sine of each element in `x`, in radians and in the
        closed interval ``[-pi/2, pi/2]``.

    See Also
    --------
    sin, cos, arccos, tan, arctan, arctan2

    Notes
    -----
    `arcsin` is a multivalued function: for each `x` there are infinitely
    many numbers `z` such that :math:`sin(z) = x`.  The convention is to
    return the angle `z` whose real part lies in [-pi/2, pi/2].

    For real-valued input data types, *arcsin* always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `arcsin` is a complex analytic function that
    has, by convention, the branch cuts [-inf, -1] and [1, inf]  and is
    continuous from above on the former and from below on the latter.

    The inverse sine is also known as `asin` or sin^{-1}.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.arcsin.html

    Abramowitz, M. and Stegun, I. A., *Handbook of Mathematical Functions*,
    10th printing, New York: Dover, 1964, pp. 79ff.
    http://www.math.sfu.ca/~cbm/aands/

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.arcsin(1)     # pi/2
    Tensor(1.5707963267948966)
    >>> mg.arcsin(-1)    # -pi/2
    Tensor(-1.5707963267948966)
    >>> mg.arcsin(0)
    Tensor(0.0)
    """
    ...


@ufunc_creator(Arccos)
def arccos(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Inverse cosine, element-wise.

    This docstring was adapted from that of numpy.arccos [1]_

    Parameters
    ----------
    x : ArrayLike
        `x`-coordinate on the unit circle.
        For real arguments, the domain is [-1, 1].

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    angle : Tensor
        The angle of the ray intersecting the unit circle at the given
        `x`-coordinate in radians [0, pi].

    See Also
    --------
    cos, arctan, arcsin

    Notes
    -----
    `arccos` is a multivalued function: for each `x` there are infinitely
    many numbers `z` such that `cos(z) = x`. The convention is to return
    the angle `z` whose real part lies in `[0, pi]`.

    For real-valued input data types, `arccos` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `arccos` is a complex analytic function that
    has branch cuts `[-inf, -1]` and `[1, inf]` and is continuous from
    above on the former and from below on the latter.

    The inverse `cos` is also known as `acos` or cos^-1.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.arccos.html

    M. Abramowitz and I.A. Stegun, "Handbook of Mathematical Functions",
    10th printing, 1964, pp. 79. http://www.math.sfu.ca/~cbm/aands/

    Examples
    --------
    We expect the arccos of 1 to be 0, and of -1 to be pi:

    >>> import mygrad as mg
    >>> mg.arccos([1, -1])
    Tensor([ 0.        ,  3.14159265])
    """
    ...


@ufunc_creator(Arctan)
def arctan(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Inverse tangent, element-wise.

    This docstring was adapted from that of numpy.arctan [1]_

    Parameters
    ----------
    x : ArrayLike

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    out : Tensor

    See Also
    --------
    arctan2 : The "four quadrant" arctan of the angle formed by (`x`, `y`)
        and the positive `x`-axis.

    Notes
    -----
    `arctan` is a multi-valued function: for each `x` there are infinitely
    many numbers `z` such that tan(`z`) = `x`.  The convention is to return
    the angle `z` whose real part lies in [-pi/2, pi/2].

    For real-valued input data types, `arctan` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `arctan` is a complex analytic function that
    has [`1j, infj`] and [`-1j, -infj`] as branch cuts, and is continuous
    from the left on the former and from the right on the latter.

    The inverse tangent is also known as `atan` or tan^{-1}.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.arctan.html

    Abramowitz, M. and Stegun, I. A., *Handbook of Mathematical Functions*,
    10th printing, New York: Dover, 1964, pp. 79.
    http://www.math.sfu.ca/~cbm/aands/

    Examples
    --------
    We expect the arctan of 0 to be 0, and of 1 to be pi/4:

    >>> import mygrad as mg
    >>> mg.arctan([0, 1])
    Tensor([ 0.        ,  0.78539816])

    >>> mg.pi / 4
    0.78539816339744828
    """
    ...


@ufunc_creator(Arctan2)
def arctan2(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Element-wise arc tangent of ``x1/x2`` choosing the quadrant correctly.

    This docstring was adapted from that of numpy.arctan [1]_

    Parameters
    ----------
    x1 : ArrayLike
        ``y``-coordinates.

    x2 : ArrayLike
        ``x``-coordinates.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    angle : Tensor
        Tensor of angles in radians, in the range ``[-pi, pi]``.

    See Also
    --------
    arctan, tan

    Notes
    -----
    *arctan2* is identical to the `atan2` function of the underlying
    C library.  The following special values are defined in the C
    standard: [2]_

    ====== ====== ================
    `x1`   `x2`   `arctan2(x1,x2)`
    ====== ====== ================
    +/- 0  +0     +/- 0
    +/- 0  -0     +/- pi
     > 0   +/-inf +0 / +pi
     < 0   +/-inf -0 / -pi
    +/-inf +inf   +/- (pi/4)
    +/-inf -inf   +/- (3*pi/4)
    ====== ====== ================

    Note that +0 and -0 are distinct floating point numbers, as are +inf
    and -inf.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.arctan.html
    .. [2] ISO/IEC standard 9899:1999, "Programming language C."

    Examples
    --------
    Consider four points in different quadrants:

    >>> import mygrad as mg
    >>> x = mg.tensor([-1.0, +1.0, +1.0, -1.0])
    >>> y = mg.tensor([-1.0, -1.0, +1.0, +1.0])
    >>> mg.arctan2(y, x) * 180 / mg.pi
    Tensor([-135.,  -45.,   45.,  135.])

    Note the order of the parameters. `arctan2` is defined also when `x2` = 0
    and at several other special points, obtaining values in
    the range ``[-pi, pi]``:

    >>> mg.arctan2([1., -1.], [0., 0.])
    Tenor([ 1.57079633, -1.57079633])
    >>> mg.arctan2([0., 0., mg.inf], [+0., -0., mg.inf])
    Tenor([ 0.        ,  3.14159265,  0.78539816])"""
    ...


def sinc(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> sin(a) / a``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Sinc, a, constant=constant)


def cot(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> cot(a)``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Cot, a, constant=constant)


def csc(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> csc(a)``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Csc, a, constant=constant)


def sec(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> sec(a)``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Sec, a, constant=constant)


def arccsc(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> arccsc(a)``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Arccsc, a, constant=constant)


def arccot(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> arccot(a)``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Arccot, a, constant=constant)


def arcsec(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """``f(a) -> arcsec(a)``

    Parameters
    ----------
    a : ArrayLike

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor"""
    return Tensor._op(Arcsec, a, constant=constant)



from typing import Optional, Sequence, Union

from mygrad.tensor_base import Tensor, implements_numpy_override
from mygrad.typing import ArrayLike

from .ops import Repeat

__all__ = ["repeat"]


@implements_numpy_override()
def repeat(
    a: ArrayLike,
    repeats: Union[int, Sequence[int]],
    axis: Optional[int] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    Repeat elements of a tensor.

    This docstring was adapted from ``numpy.repeat``

    Parameters
    ----------
    a : ArrayLike
        Input tensor.

    repeats : Union[int, Sequence[int]]
        The number of repetitions for each element. ``repeats``
        is broadcasted to fit the shape of the given axis.

    axis : Optional[int]
        The axis along which to repeat values. By default, use the
        flattened input array, and return a flat output tensor.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.
    Returns
    -------
    repeated_tensor : Tensor
        Output tensor which has the same shape as `a`, except along
        the given axis.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.repeat(3, 4)
    Tensor([3, 3, 3, 3])
    >>> x = mg.Tensor([[1, 2], [3, 4]])
    >>> mg.repeat(x, 2)
    Tensor([1, 1, 2, 2, 3, 3, 4, 4])
    >>> mg.repeat(x, 3, axis=1)
    Tensor([[1, 1, 1, 2, 2, 2],
            [3, 3, 3, 4, 4, 4]])
    >>> mg.repeat(x, [1, 2], axis=0)
    Tensor([[1, 2],
            [3, 4],
            [3, 4]])
    """
    return Tensor._op(Repeat, a, op_args=(repeats, axis), constant=constant)



from typing import TYPE_CHECKING, Type, TypeVar, Union

from numpy import dtype

if TYPE_CHECKING:  # pragma: no cover
    DTypeLike = Union[dtype, None, type, str]
else:  # pragma: no cover
    DTypeLike = TypeVar(
        "DTypeLike",
        bound=Union[dtype, None, type, str],
    )

if TYPE_CHECKING:  # pragma: no cover
    DTypeLikeReals = Union[dtype, None, Type[bool], Type[int], Type[float], str]

else:  # pragma: no cover
    DTypeLikeReals = TypeVar(
        "DTypeLikeReals",
        bound=Union[dtype, None, Type[bool], Type[int], Type[float], str],
    )



from numbers import Real
from typing import Optional

from numpy import ndarray

from mygrad import maximum, minimum
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike

__all__ = ["leaky_relu"]


def leaky_relu(
    x: ArrayLike, slope: float, *, constant: Optional[bool] = None
) -> Tensor:
    """Returns the leaky rectified linear activation elementwise along x.

    The leaky ReLU is given by `max(x, 0) + slope*min(x, 0)`.

    Parameters
    ----------
    x : ArrayLike
        Input data.

    slope : Union[Real, mygrad.Tensor]
        The slope of the negative activation.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor
        The result of apply the "leaky relu" function elementwise to `x`.

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet.activations import leaky_relu
    >>> x = mg.arange(-5, 6)
    >>> x
    Tensor([-5, -4, -3, -2, -1,  0,  1,  2,  3,  4,  5])
    >>> y = leaky_relu(x, slope=0.1); y
    >>> Tensor([-0.5, -0.4, -0.3, -0.2, -0.1,  0. ,  1. ,  2. ,  3. ,  4. ,  5. ])
    >>> y.backward()
    >>> x.grad
    array([0.1, 0.1, 0.1, 0.1, 0.1, 0. , 1. , 1. , 1. , 1. , 1. ])

    .. plot::

       >>> import mygrad as mg
       >>> from mygrad.nnet.activations import leaky_relu
       >>> import matplotlib.pyplot as plt
       >>> x = mg.linspace(-2, 2, 100)
       >>> y = leaky_relu(x, slope=0.1)
       >>> plt.title("leaky_relu(x, slope=0.1)")
       >>> y.backward()
       >>> plt.plot(x, x.grad, label="df/dx")
       >>> plt.plot(x, y, label="f(x)")
       >>> plt.legend()
       >>> plt.grid()
       >>> plt.show()
    """
    if isinstance(slope, (ndarray, Tensor)):
        slope = slope.item()

    if not isinstance(slope, Real):
        raise TypeError(
            f"`slope` must be a real-valued scalar, got {slope} (type { type(slope)})"
        )

    return maximum(x, 0, constant=constant) + slope * minimum(x, 0, constant=constant)



"""
Provides user interface for suspending computational graph tracking and back-propagation
"""
from functools import wraps
from typing import Callable, TypeVar, Any, cast, Optional

import numpy as np

from mygrad._utils import ContextTracker

__all__ = ["no_autodiff"]

_T = TypeVar("_T", bound=Callable[..., Any])

# If `False`, suspends all computational graph tracking and backprop
TRACK_GRAPH = True  # type: bool


class _NoAutoDiff(ContextTracker):
    """Serves as a context manager and decorator for suspending
    all computational graph tracking.

    Note that memory guarding does not occur in the `no_autodiff` context,
    so there is no need to nest this context with `mem_guard_off`.

    Examples
    --------
    Demonstrating ``no_autodiff`` as a context-manager

    >>> import mygrad as mg
    >>> with mg.no_autodiff:
    >>>     # all computational graph tracking is suspended
    >>>     # within the context
    >>>     x = mg.arange(4.)
    >>>     (4 * x).backward()  # no autodiff will occur
    >>> x.grad is None


    Demonstrating ``no_autodiff`` as a decorator

    >>> @mg.no_autodiff
    ... def func():
    ...     # No graph-tracking will occur within
    ...     # the body of this function
    ...     pass
    """

    _enter_set_value = False

    @property
    def state(self):
        return TRACK_GRAPH

    @state.setter
    def state(self, value: bool):
        if not isinstance(value, bool):  # pragma: no cover
            raise TypeError(
                f"TRACK_GRAPH must be set to a boolean value, got {value} (type={type(value)})"
            )

        global TRACK_GRAPH
        TRACK_GRAPH = value

    def __call__(self, func: _T, to_numpy: bool = False) -> _T:
        """Decorates a function so that it will have graph-tracking suspended
        during its execution.

        Parameters
        ----------
        func : Callable
            The function to be decorated

        to_numpy : bool, optional (default=False)
            If true, the output is assumed to be array-like and
            will be cast to a numpy array

        Returns
        -------
        decorated_func : Callable"""

        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                out = func(*args, **kwargs)
            return out if not to_numpy else np.asarray(out)

        return cast(_T, wrapper)


no_autodiff = _NoAutoDiff()



class MyGradException(Exception):
    """Generic parent class for exceptions thrown by MyGrad."""


class InvalidGradient(MyGradException):
    """An invalid gradient (i.e. a non-numeric or non-array-like object)
    was produced by an operation or was supplied by a user."""


class InvalidBackprop(MyGradException):
    """Backpropagation was invoked through a partially-cleared graph
    or from a non-scalar for a scalar-only graph"""


class DisconnectedView(MyGradException):
    custom_msg = (
        "An inplace operation was invoked on a tensor-view that "
        "is not connected to its base tensor through a computational "
        "graph."
        "It is likely that this tensor was involved during backpropagation;"
        "consider recreating this view before proceeding."
    )



from numbers import Integral

import numpy as np
from numpy.lib.stride_tricks import as_strided


def sliding_window_view(arr, window_shape, step, dilation=None):
    """Create a sliding window view over the trailing dimensions of an array.
    No copy is made unless the input array is not contiguous in memory.

    The window is applied only to valid regions of ``arr``, but is applied greedily.

    See Notes section for details.

    Parameters
    ----------
    arr : numpy.ndarray, shape=(..., [x, (...), z])
        C-contiguous array over which sliding view-window is applied along the trailing
        dimensions ``[x, ..., z]``, as determined by the length of ``window_shape``.

        If ``arr`` is not C-contiguous, it will be replaced by ``numpy.ascontiguousarray(arr)``

    window_shape : Sequence[int]
        Specifies the shape of the view-window: ``[Wx, (...), Wz]``.
        The length of `window_shape` determines the length of ``[x, (...) , z]``.

    step : Union[int, Sequence[int]]
        The step sized used along the ``[x, (...), z]`` dimensions: ``[Sx, (...), Sz]``.
        If a single integer is specified, a uniform step size is used.

    dilation : Optional[Union[int, Sequence[int]]]
        The dilation factor used along the ``[x, (...), z]`` directions: ``[Dx, (...), Dz]``.
        If no value is specified, a dilation factor of 1 is used along each direction.
        Dilation specifies the step size used when filling the window's elements

    Returns
    -------
    numpy.ndarray
        A contiguous view of ``arr``, of shape ``([X, (...), Z], ..., [Wx, (...), Wz])``, where
        ``[X, ..., Z]`` is the shape of the grid on which the window was applied. See Notes
        sections for more details.

    Raises
    ------
    ValueError, TypeError
        Invalid step-size, window shape, or dilation

    Notes
    -----
    Window placement:
        Given a dimension of size x, with a window of size W along this dimension, applied
        with stride S and dilation D, the window will be applied::
                                  X = (x - (W - 1) * D + 1) // S + 1
        number of times along that dimension.

    Interpreting output:
        In general, given an array ``arr`` of shape (..., x, (...), z), and::

            out = sliding_window_view(arr, window_shape=[Wx, (...), Wz], step=[Sx, (...), Sz])

        then indexing ``out`` with ``[ix, (...), iz]`` produces the following view of ``x``::

            out[ix, (...), iz] ==
                x[..., ix*Sx:(ix*Sx + Wx*Dx):Dx, (...), iz*Sz:(iz*Sz + Wz*Dz):Dz]

        For example, suppose ``arr`` is an array of shape-(10, 12, 6). Specifying sliding
        window of shape ``(3, 3)`` with step size ``(2, 2)``, dilation ``(2, 1)`` will create the view::

                        [[arr[:,  0:6:2, 0:3], arr[:,   0:6:3, 3:6]]
                         [arr[:, 6:12:2, 0:3], arr[:, 6:12:12, 3:6]]]

        producing a view of shape ``(2, 2, 10, 3, 3)`` in total.

    Examples
    --------
    >>> import numpy as np
    >>> x = np.arange(36).reshape(6, 6)
    >>> x
    array([[ 0,  1,  2,  3,  4,  5],
           [ 6,  7,  8,  9, 10, 11],
           [12, 13, 14, 15, 16, 17],
           [18, 19, 20, 21, 22, 23],
           [24, 25, 26, 27, 28, 29],
           [30, 31, 32, 33, 34, 35]])

    Apply an 3x2 window with step-sizes of (2, 2). This results in
    the window being placed twice along axis-0 and three times along axis-1.

    >>> y = sliding_window_view(x, step=(2, 2), window_shape=(3, 2))
    >>> y.shape
    (2, 3, 3, 2)

    window applied at (0, 0)

    >>> y[0, 0]
    array([[ 0,  1],
           [ 6,  7],
           [12, 13]])

    window applied at (2, 0)

    >>> y[1, 0]
    array([[12, 13],
           [18, 19],
           [24, 25]])

    window applied at (0, 2)

    >>> y[0, 1]
    array([[ 2,  3],
           [ 8,  9],
           [14, 15]])

    verify that an element in this window-view is correct

    >>> i, j = np.random.randint(0, 2, size=2)
    >>> wx, wy = (2, 2)
    >>> sx, sy = (2, 2)
    >>> np.all(y[i, j] == x[..., i*sx:(i*sx + wx), j*sy:(j*sy + wy)])
    True
    """

    if not hasattr(window_shape, "__iter__"):
        raise TypeError(
            f"`window_shape` must be a sequence of positive integers, got: {window_shape}"
        )
    window_shape = tuple(window_shape)
    if not all(isinstance(i, Integral) and i > 0 for i in window_shape):
        raise TypeError(
            f"`window_shape` must be a sequence of positive integers, "
            f"got: {window_shape}"
        )

    if len(window_shape) > arr.ndim:
        raise ValueError(
            f"`window_shape` ({window_shape}) cannot specify more values than "
            f"`arr.ndim` ({arr.ndim})."
        )

    if not isinstance(step, Integral) and not hasattr(step, "__iter__"):
        raise TypeError(
            f"`step` must be a positive integer or a sequence of positive "
            f"integers, got: {step}"
        )

    step = (
        (int(step),) * len(window_shape) if isinstance(step, Integral) else tuple(step)
    )

    if not all(isinstance(i, Integral) and i > 0 for i in step):
        raise ValueError(
            f"`step` must be a positive integer or a sequence of positive "
            f"integers, got: {step}"
        )

    if any(i > j for i, j in zip(window_shape[::-1], arr.shape[::-1])):
        raise ValueError(
            f"Each size of the window-shape must fit within the trailing "
            f"dimensions of `arr`."
            f"{window_shape} does not fit in {arr.shape[-len(window_shape) :]}"
        )

    if (
        dilation is not None
        and not isinstance(dilation, Integral)
        and not hasattr(dilation, "__iter__")
    ):
        raise TypeError(
            f"`dilation` must be None, a positive integer, or a sequence of "
            f"positive integers, got: {dilation}"
        )
    if dilation is None:
        dilation = np.ones((len(window_shape),), dtype=int)
    else:
        if isinstance(dilation, Integral):
            dilation = np.full((len(window_shape),), fill_value=dilation, dtype=int)
        else:
            np.asarray(dilation)

        if not all(isinstance(i, Integral) and i > 0 for i in dilation) or len(
            dilation
        ) != len(window_shape):
            raise ValueError(
                f"`dilation` must be None, a positive integer, or a sequence of "
                f"positive integers with the same length as `window_shape` "
                f"({window_shape}), got: {dilation}"
            )
        if any(
            w * d > s
            for w, d, s in zip(window_shape[::-1], dilation[::-1], arr.shape[::-1])
        ):
            raise ValueError(
                f"The dilated window ({tuple(w * d for w, d in zip(window_shape, dilation))}) "
                f"must fit within the trailing "
                f"dimensions of `arr` ({arr.shape[-len(window_shape) :]})"
            )

    if not arr.flags["C_CONTIGUOUS"]:
        arr = np.ascontiguousarray(arr)

    step = np.array(step)  # (Sx, ..., Sz)
    window_shape = np.array(window_shape)  # (Wx, ..., Wz)
    in_shape = np.array(arr.shape[-len(step) :])  # (x, ... , z)
    nbyte = arr.strides[-1]  # size, in bytes, of element in `arr`

    # per-byte strides required to fill a window
    win_stride = tuple(np.cumprod(arr.shape[:0:-1])[::-1]) + (1,)

    # per-byte strides required to advance the window
    step_stride = tuple(win_stride[-len(step) :] * step)

    # update win_stride to accommodate dilation
    win_stride = np.array(win_stride)
    win_stride[-len(step) :] *= dilation
    win_stride = tuple(win_stride)

    # tuple of bytes to step to traverse corresponding dimensions of view
    # see: 'internal memory layout of an ndarray'
    stride = tuple(int(nbyte * i) for i in step_stride + win_stride)

    # number of window placements along x-dim: X = (x - (Wx - 1)*Dx + 1) // Sx + 1
    out_shape = tuple((in_shape - ((window_shape - 1) * dilation + 1)) // step + 1)

    # ([X, (...), Z], ..., [Wx, (...), Wz])
    out_shape = out_shape + arr.shape[: -len(step)] + tuple(window_shape)
    out_shape = tuple(int(i) for i in out_shape)

    return as_strided(arr, shape=out_shape, strides=stride, writeable=False)



from abc import ABC, abstractmethod
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Optional,
    Set,
    TypeVar,
    Union,
)
from weakref import ReferenceType

if TYPE_CHECKING:  # pragma: no cover
    from mygrad import Tensor
    from mygrad.operation_base import Operation

__all__ = [
    "collect_all_operations_and_clear_grads",
    "ContextTracker",
    "reduce_broadcast",
    "SkipGradient",
    "WeakRef",
    "WeakRefIterable",
]


T = TypeVar("T")


def collect_all_operations_and_clear_grads(
    t: "Tensor", seen: Set["WeakRef[Operation]"]
):
    """Recursively accumulates in `seen` all operations involved
    in creating `t`.

    `seen` is updated in-place
    """
    t._view_grad = None
    t._grad = None

    if t.creator is None or t.constant:
        return

    c = ReferenceType(t.creator)  # type: WeakRef[Operation]

    if c in seen:
        return

    seen.add(c)

    for t in t.creator.variables:
        collect_all_operations_and_clear_grads(t, seen)


class WeakRef(Generic[T]):
    __slots__ = ()

    def __init__(self, ob: T, callback=None, **annotations):  # pragma: no cover
        ...

    def __call__(self) -> Union[None, T]:  # pragma: no cover
        ...


class WeakRefIterable(Generic[T]):
    """
    Stores weakrefs in a list and, upon iteration, yields
    only living references.
    """

    __slots__ = ("data",)

    def __init__(self, data: Optional[Iterable[T]] = None):
        self.data: List[WeakRef[T]] = []

        if data is not None:
            self.data: List[WeakRef[T]] = list(ReferenceType(x) for x in data)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __bool__(self):
        if len(self.data) == 0:
            return False
        else:
            return any(True for _ in self)

    def append(self, item: T):
        self.data.append(ReferenceType(item))

    def clear(self):
        self.data.clear()

    def __iter__(self) -> Generator[T, None, None]:
        for ref in self.data:
            item = ref.__call__()  # use __call__ to help type checker..
            if item is not None:
                yield item


class SkipGradient(Exception):
    """The gradient for the current tensor-label pair has already
    been computed, scaled, and back-propped, skip gradient calculation."""


def reduce_broadcast(grad, var_shape):
    """Sum-reduce axes of `grad` so its shape matches `var_shape.

    This the appropriate mechanism for backpropagating a gradient
    through an operation in which broadcasting occurred for the
    given variable.

    Parameters
    ----------
    grad : numpy.ndarray
    var_shape : Tuple[int, ...]

    Returns
    -------
    numpy.ndarray

    Raises
    ------
    ValueError
        The dimensionality of the gradient cannot be less than
        that of its associated variable."""
    if grad.shape == var_shape:
        return grad

    if grad.ndim != len(var_shape):
        if grad.ndim < len(var_shape):
            raise ValueError(
                f"The dimensionality of the gradient of the broadcasted "
                f"operation ({grad.ndim}) is less than that of its associated variable "
                f"({len(var_shape)})"
            )
        grad = grad.sum(axis=tuple(range(grad.ndim - len(var_shape))))

    keepdims = tuple(n for n, i in enumerate(grad.shape) if i != var_shape[n])
    if keepdims:
        grad = grad.sum(axis=keepdims, keepdims=True)

    return grad


class ContextTracker(ABC):
    """A context manager and decorator for managing a boolean
    global state"""

    # tracks context depth
    _depth = 0  # type: int

    # the value that the state is set to upon entering the context
    _enter_set_value: Optional[bool] = None

    @property
    @abstractmethod
    def state(self) -> bool:  # pragma: no cover
        raise NotImplementedError()

    @state.setter
    @abstractmethod
    def state(self, value: bool):  # pragma: no cover
        raise NotImplementedError()

    def __init__(self):
        # keeps track of what MemGuard was at a given depth
        self._depth_tracker: Dict[int, bool] = dict()

    def __bool__(self) -> bool:  # pragma: no cover
        return self.state

    def __enter__(self):
        self._depth_tracker[self._depth] = self.state
        self._depth += 1
        self.state = self._enter_set_value

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._depth -= 1
        self.state = self._depth_tracker.pop(self._depth)

    def __call__(self, func: Callable) -> Callable:
        """Decorates a function within the context"""

        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper



from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Union,
)

from numpy import logical_not, ndarray

from mygrad._utils import WeakRefIterable
from mygrad.operation_base import Operation

if TYPE_CHECKING:  # pragma: no cover
    from mygrad import Tensor


T = TypeVar("T")


def mirror_tensor(*, target: "Tensor", source: "Tensor"):
    """*Dev use only*

    Creates a shallow copy of attribute dictionary of ``self`` and assigns
    it to``tensor``, so that ``tensor`` has the same state as ``self`` and
    points to the same array data.

    This is used to facilitate "in-place" operations.
    """
    target.__dict__ = source.__dict__.copy()


def reroute_ops_through(*, target: "Tensor", source: "Tensor"):
    for op in source._ops:
        op = op()
        if op is None:
            continue

        op.variables = tuple(
            var_ if var_ is not source else target for var_ in op.variables
        )


def make_placeholder_tensor(
    original: "Tensor", *, base: Optional["Tensor"] = None
) -> "Tensor":
    """
    Creates a tensor that stands in the place of `original` in the computational graph.
    This does not create a copy of the array-data held by original; the mirrored tensor
    points to the same data.

    The resulting tensor should never be exposed to the user; it is used to accommodate
    in-place operations.

    Parameters
    ----------
    original : bool
        If True the placeholder holds a copy of the original data

    base : Optional[Tensor]
        Points the placeholder to the base tensor.

    Returns
    -------
    placeholder : Tensor
    """
    assert (
        original._grad is None
    ), "A placeholder copy can not be created for a tensor with a gradient"

    assert (
        not original._accum_ops
    ), "A placeholder copy cannot be created during backprop"

    placeholder = type(original)([])
    mirror_tensor(target=placeholder, source=original)
    placeholder._base = base
    # point all ops involving `self` to old_tensor instead
    reroute_ops_through(target=placeholder, source=original)
    return placeholder


class Node(NamedTuple):
    tensor: "Tensor"
    placeholder: "Tensor"
    # `tensor` is a view of `parent`
    parent: Optional["Tensor"] = None


class DuplicatingGraph:
    """Traces through the graph of all views of the base tensor and
    creates a corresponding graph of 'placeholders', used to permit
    a future in-place operation without mutating the current tensor
    graph.

    Provides the information needed to recreate a view-graph after
    an in-place operation has been performed on the base tensor.

    Upon initialization, this class mutates the graph downstream of the
    base tensor.
    """

    def _duplicate_graph(self, tensor: "Tensor"):
        """Recursively creates placeholders for all views downstream of `tensor`.

        Upon completion, the existing computational graph involves only placeholders.
        Note that the placeholders and original tensors point to the same array-data."""
        if not tensor._view_children:
            self.leafs.add(id(tensor))
            return

        for child in tensor._view_children:

            self._record_mapping(
                original=child,
                placeholder=make_placeholder_tensor(
                    original=child, base=self.base.placeholder
                ),
                parent=tensor,
            )

            self._duplicate_graph(child)

        self[tensor].placeholder._view_children = WeakRefIterable(
            [self[t].placeholder for t in tensor._view_children]
        )

    def __init__(self, base: "Tensor"):
        self.mappings: Dict[int, Node] = {}

        self._record_mapping(
            original=base, placeholder=make_placeholder_tensor(base, base=base.base)
        )
        self.base = self[base]

        self.leafs: Set[int] = set()
        # creates placeholders for each node in the view graph
        self._duplicate_graph(base)

    def __getitem__(self, item: "Tensor") -> Node:
        """Returns a node associated with a tensor"""
        return self.mappings[id(item)]

    def _record_mapping(
        self,
        original: "Tensor",
        placeholder: "Tensor",
        parent: Optional["Tensor"] = None,
    ):
        """
        Parameters
        ----------
        original : Tensor
            A tensor that will be involved in a mutated graph

        placeholder : Tensor
            Takes the place of the original in the computational graph

        parent : Optional[Tensor]
            The tensor of which ``original`` is a direct view
        """
        node = Node(tensor=original, placeholder=placeholder, parent=parent)
        self.mappings[id(node.tensor)] = node
        self.mappings[id(node.placeholder)] = node

    def _yield_children(self, tensor: "Tensor") -> Iterator[Node]:
        """Recursive helper function for DFS iteration"""
        yield self[tensor]
        for child in tensor._view_children:
            yield from self._yield_children(child)

    def __contains__(self, item):
        return id(item) in self.mappings

    def get_placeholder_if_exists(self, tensor: T) -> T:
        if tensor in self:
            return self[tensor].placeholder
        else:
            return tensor

    def __iter__(self) -> Iterator[Node]:
        """Returns all nodes in graph using DFS.

        Note that each node can only have one input edge, so no visitor
        information need be recorded

        We iterate based off of the placeholders' graph information
        since they will never be mutated."""
        yield from self._yield_children(self.base.placeholder)

    def get_path_to_base(self, tensor: "Tensor") -> List[Node]:
        """ Returns [leaf, (parent), ..., base]"""
        path = []
        node = self[tensor]
        while node.parent is not None:
            path.append(node)
            node = self[node.parent]
        path.append(self.base)
        return path

    def restore_old_graph(self):
        """ Reroute graph back to original tensors."""
        # call tuple to ensure iteration is completed
        # before information gets deleted / mutated
        for node in tuple(self):
            reroute_ops_through(target=node.tensor, source=node.placeholder)
            if node.placeholder._base is not None:
                node.tensor._base = self.base.tensor


class UnView(Operation):
    """
    Creates an operation that connects a mutant base to
    the placeholder mutant-view and placeholder base that
    it is derived from.

    This effectively connects the mutant base to the upstream
    computational graph.
    """

    def __call__(
        self,
        placeholder_base: "Tensor",
        placeholder_mutant_view: "Tensor",
        mutant_base_data: ndarray,
        view_fn_sequence: Sequence[Callable[[ndarray], ndarray]],
    ):
        """
        Parameters
        ----------
        placeholder_mutant_view: Tensor
            The internal tensor that resulted from the in-place
            operation.

        placeholder_base: Tensor
            The placeholder for the base tensor involved in the
            in-place operation.

        mutant_base_data: ndarray
            The base tensor that was mutated by the in-place operation,
            and that will be exposed to the user.

        view_fn_sequence: Sequence[Callable[[ndarray], ndarray]]
            The sequence of view-functions used to create the
            view-tensor from the base

        Returns
        -------
        mutant_base_data : ndarray
            The array associated with the mutant base
        """
        self.variables = (placeholder_base, placeholder_mutant_view)
        self._view_fn_seq = view_fn_sequence
        return mutant_base_data

    def backward_var(self, grad: ndarray, index: int, **kwargs) -> ndarray:
        placeholder_base, placeholder_mutant_view = self.variables

        # Backprop through upstream base by zeroing out
        # all regions of `grad` associated with the downstream
        # view.
        #
        # Backprop through upstream view by taking the corresponding
        # view of the gradient
        #
        # E.g.
        #
        # base -> [1., 2., 3.]
        # view = base[:2] -> [1., 2.]
        # dℒ/d(out) = [g0, g1, g2]
        #
        # dℒ/d(base) = [0., 0., g2]
        # dℒ/d(view) = [g0, g1]
        if index == 0:  # compute dℒ/d(base)
            grad = grad.copy()
            grad_view = grad
            for fn in self._view_fn_seq:
                grad_view = fn(grad_view)

            assert grad_view.shape == self.variables[1].shape
            # check that grad_view shares memory with grad
            assert grad_view.base is grad
            grad_view *= 0

            return grad

        elif index == 1:  # compute dℒ/d(view)
            grad_view = grad
            for fn in self._view_fn_seq:
                grad_view = fn(grad_view)

            assert grad_view.shape == self.variables[1].shape
            return grad_view

        else:  # pragma: no cover
            raise ValueError(f"UnView: backward_var index: {index}")


class ApplyMask(Operation):
    """When an operation like `multiply(x, y, where=mask, out=z)` occurs,
    we need to connect the upstream version of `z` to the computational
    graph so that `~mask * dℒ/dz` backprops to it, whereas `~mask * dℒ/dz`
    will backprop to `x` and `y`.

                      old-z ---------------------
                        |                       |
            multiply(x, y, where=mask, out=z)   |
                        |                       |
                        z    --------------------
                        |    |
                       ApplyMask
                          |
                          z (unchanged in value by ApplyMask, but now backprops to `old-z`

    This is basically an alternative to treating `multiply(x, y, where=mask, out=z)`
    like a three-input operation, would add complexity to the implementation of
    every op that supports `where` and `out`. So instead, we delegate this rote
    task to be performed by `Tensor._inplace_op`
    """

    def __call__(
        self,
        pass_through_tensor: "Tensor",
        upstream_tensor: "Tensor",
        *,
        mask: Union[bool, ndarray],
    ) -> ndarray:
        self.variables = (pass_through_tensor, upstream_tensor)
        self._mask = mask
        return pass_through_tensor.data

    def backward_var(self, grad: ndarray, index: int, **kwargs) -> ndarray:
        if index == 0:
            return grad
        elif index == 1:
            return grad * logical_not(self._mask)
        else:  # pragma: no cover
            raise ValueError(f"UnView: backward_var index: {index}")



from numbers import Real
from typing import Optional, Sequence, Tuple, Union

import numpy as np
from numpy.core.einsumfunc import _parse_einsum_input

import mygrad as mg
from mygrad.math.misc.funcs import absolute
from mygrad.math.sequential.funcs import max as mg_max, min as mg_min
from mygrad.tensor_base import Tensor, implements_numpy_override
from mygrad.typing import ArrayLike

from .ops import EinSum, Norm

__all__ = ["einsum", "norm"]


@implements_numpy_override(np.linalg.norm)
def norm(
    x: ArrayLike,
    ord: Optional[Union[int, float]] = None,
    axis: Optional[Union[int, Tuple[int]]] = None,
    keepdims: bool = False,
    *,
    nan_to_num: bool = True,
    constant: Optional[bool] = None,
) -> Tensor:
    r"""Vector norm.

    This function is an infinite number of vector norms (described below), depending
    on the value of the ``ord`` parameter.

    In contrast to ``numpy.linalg.norm``, matrix norms are not supported.

    This docstring was adapted from that of ``numpy.linalg.norm`` [1]_.

    Parameters
    ----------
    x : ArrayLike
        Input tensor.  If `axis` is None, then `x` must be 1-D unless `ord`
        is None. If both `axis` and `ord` are None, the 2-norm of
        ``x.ravel`` will be returned.

    ord : Optional[Union[int, float]]
        Order of the norm (see table under ``Notes``). inf means numpy's
        `inf` object. The default is None.

    axis : Optional[Union[int, Tuple[int]]]
        If `axis` is an integer, it specifies the axis of `x` along which to
        compute the vector norms. The default is None.

    keepdims : bool, optional (default=False)
        If this is set to True, the axes which are normed over are left in the
        result as dimensions with size one.  With this option the result will
        broadcast correctly against the original `x`.

    nan_to_num : bool, optional (default=True)
        If `True` then gradients that would store nans due to the presence of
        zeros in `x` will instead store zeros in those places.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        Norm(s) of the vector(s).

    Notes
    -----
    For values of ``ord < 1``, the result is, strictly speaking, not a
    mathematical 'norm', but it may still be useful for various numerical
    purposes.

    The following norms can be calculated:

    =====  ==========================
    ord    norm for vectors
    =====  ==========================
    inf    max(abs(x))
    -inf   min(abs(x))
    0      sum(x != 0)
    1      as below
    -1     as below
    2      as below
    -2     as below
    other  sum(abs(x)**ord)**(1./ord)
    =====  ==========================

    The Frobenius norm is given by [1]_:

        :math:`||A||_F = [\sum_{i,j} abs(a_{i,j})^2]^{1/2}`

    The nuclear norm is the sum of the singular values.

    Both the Frobenius and nuclear norm orders are only defined for
    matrices and raise a ValueError when ``x.ndim != 2``.

    References
    ----------
    .. [1] Retrived from: https://numpy.org/doc/stable/reference/generated/numpy.linalg.norm.html
    .. [2] G. H. Golub and C. F. Van Loan, *Matrix Computations*,
           Baltimore, MD, Johns Hopkins University Press, 1985, pg. 15

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.tensor([[1.0, 2.0, 3.0],
    ...                [1.0, 0.0, 0.0]])
    >>> l2_norms = mg.linalg.norm(x, axis=1, ord=2)
    >>> l2_norms
    Tensor([3.74165739, 1.        ])

    The presence of the elementwise absolute values in the norm operation means that zero-valued entries in any of 
    input vectors have an undefined derivative. When `nan_to_num=False` is specified these derivatives will be reported
    as `nan`, otherwise they will be made to be 0.0.

    >>> l2_norms = mg.linalg.norm(x, axis=1, ord=2, nan_to_num=True)
    >>> l2_norms.backward()
    >>> x.grad
    array([[0.26726124, 0.53452248, 0.80178373],
           [1.        ,        nan,        nan]])

    This is rigorously true, but is often not the desired behavior in autodiff applications.
    Rather, it can be preferable to use `0.0` to fill these undefined derivatives.
    This is the default behavior, when `nan_to_num` is not specified.

    >>> l2_norms = mg.linalg.norm(x, axis=1, ord=2, nan_to_num=False)  # default setting: `nan_to_num=False`
    >>> l2_norms.backward()
    >>> x.grad
    array([[0.26726124, 0.53452248, 0.80178373],
          [1.        ,          0.,         0.]])

    L1 norms along each of the three columns:

    >>> mg.linalg.norm(x, axis=0, ord=1)
    Tensor([2., 2., 3.])
    """
    if isinstance(ord, Real) and np.isinf(ord):
        op = mg_max if ord > 0 else mg_min
        abs_ = absolute(x, constant=constant)
        out = op(abs_, axis=axis, keepdims=keepdims)

        in_ndim = abs_.creator.variables[0].ndim

        if (axis is None and ord is not None and in_ndim == 2) or (
            hasattr(axis, "__len__") and len(axis) > 1
        ):
            raise NotImplementedError(
                "mygrad.linalg.norm does not support matrix norms"
            )
        return out
    return Tensor._op(
        Norm,
        x,
        op_kwargs={
            "axis": axis,
            "keepdims": keepdims,
            "ord": ord,
            "nan_to_num": nan_to_num,
        },
        constant=constant,
    )


@implements_numpy_override()
def einsum(
    *operands: Union[ArrayLike, str, Sequence[int]],
    optimize: bool = False,
    out: Optional[Union[np.ndarray, Tensor]] = None,
    constant: Optional[bool] = None,
) -> Tensor:
    r"""
    einsum(subscripts, *operands)

    Evaluates the Einstein summation convention on the operands. This implementation
    exactly mirrors that of ``numpy.einsum`` and supports back-propagation through
    all variety of tensor-products, sums, traces, and views that it can perform.

    The following docstring was adapted from the documentation for ``numpy.einsum``

    Using the Einstein summation convention, many common multi-dimensional
    array operations can be represented in a simple fashion.  This function
    provides a way to compute such summations. The best way to understand this
    function is to try the examples below, which show how many common NumPy/MyGrad
    functions can be implemented as calls to ``einsum``.

    Back-propagation via ``einsum`` is optimized such that any tensor that occurs
    redundantly within the summation will only have its gradient computed once.
    This optimization accommodates all number and combination of redundancies that can
    be encountered.

    E.g. back-propping through ``einsum('...,...->', x, x)`` will only incur a single
    computation/accumulation for ``x.grad`` rather than two. This permits users to
    leverage the efficiency of sum-reduction, where ``(x ** 2).sum()`` is sub-optimal,
    without being penalized during back-propagation.

    Parameters
    ----------
    subscripts : str
        Specifies the subscripts for summation.

    operands : array_like
        The tensors used in the summation.

    optimize : {False, True, 'greedy', 'optimal'}, optional (default=False)
        Controls if intermediate optimization should occur; also enables
        the use of BLAS where possible. This can produce significant speedups
        for computations like matrix multiplication.

        No optimization will occur if False and True will default to the 'greedy'
        algorithm. Also accepts an explicit contraction list from the
        ``np.einsum_path`` function. See ``np.einsum_path`` for more details.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    output : mygrad.Tensor
        The calculation based on the Einstein summation convention.

    Notes
    -----
    The subscripts string is a comma-separated list of subscript labels,
    where each label refers to a dimension of the corresponding operand.
    Repeated subscripts labels in one operand take the diagonal.  For example,
    ``einsum('ii', a)`` is equivalent to ``np.trace(a)`` (however, the former
    supports back-propagation).

    Whenever a label is repeated, it is summed, so ``einsum('i, i', a, b)``
    is equivalent to ``np.inner(a, b)``.  If a label appears only once,
    it is not summed, so ``einsum('i', a)`` produces a view of ``a``
    with no changes.

    The order of labels in the output is by default alphabetical.  This
    means that ``np.einsum('ij', a)`` doesn't affect a 2D tensor, while
    ``einsum('ji', a)`` takes its transpose.

    The output can be controlled by specifying output subscript labels
    as well.  This specifies the label order, and allows summing to
    be disallowed or forced when desired.  The call ``einsum('i->', a)``
    is like ``np.sum(a, axis=-1)``, and ``einsum('ii->i', a)``
    is like ``np.diag(a)``.  The difference is that `einsum` does not
    allow broadcasting by default.

    To enable and control broadcasting, use an ellipsis.  Default
    NumPy-style broadcasting is done by adding an ellipsis
    to the left of each term, like ``einsum('...ii->...i', a)``.
    To take the trace along the first and last axes,
    you can do ``einsum('i...i', a)``, or to do a matrix-matrix
    product with the left-most indices instead of rightmost, you can do
    ``einsum('ij...,jk...->ik...', a, b)``.

    When there is only one operand, no axes are summed, and no output
    parameter is provided, a view into the operand is returned instead
    of a new tensor.  Thus, taking the diagonal as ``einsum('ii->i', a)``
    produces a view.

    An alternative way to provide the subscripts and operands is as
    ``einsum(op0, sublist0, op1, sublist1, ..., [sublistout])``. The examples
    below have corresponding `einsum` calls with the two parameter methods.

    Examples
    --------
    >>> import mygrad as mg
    >>> import numpy as np
    >>> a = mg.arange(25).reshape(5,5)
    >>> b = mg.arange(5)
    >>> c = mg.arange(6).reshape(2,3)

    Compute the trace of ``a``, :math:`\sum_{i}{A_{ii}} = f`:

    >>> einsum('ii', a)
    Tensor(60)
    >>> einsum(a, [0, 0])
    Tensor(60)
    >>> np.trace(a.data)
    array(60)

    Return a view along the diagonal of ``a``, :math:`A_{ii} = F_{i}`:

    >>> einsum('ii->i', a)
    Tensor([ 0,  6, 12, 18, 24])
    >>> einsum(a, [0,0], [0])
    Tensor([ 0,  6, 12, 18, 24])
    >>> np.diag(a.data)
    array([ 0,  6, 12, 18, 24])

    Compute the matrix-vector product of ``a`` with ``b``, :math:`\sum_{j}{A_{ij} B_{j}} = F_{i}`:

    >>> einsum('ij,j', a, b)
    Tensor([ 30,  80, 130, 180, 230])
    >>> einsum(a, [0,1], b, [1])
    Tensor([ 30,  80, 130, 180, 230])
    >>> mg.matmul(a, b)
    Tensor([ 30,  80, 130, 180, 230])
    >>> einsum('...j,j', a, b)
    Tensor([ 30,  80, 130, 180, 230])

    Take the transpose of ``c``, :math:`C_{ji} = F_{ij}`:

    >>> einsum('ji', c)
    Tensor([[0, 3],
            [1, 4],
            [2, 5]])
    >>> einsum(c, [1, 0])
    Tensor([[0, 3],
            [1, 4],
            [2, 5]])
    >>> c.T
    Tensor([[0, 3],
            [1, 4],
            [2, 5]])

    Compute ``3 * c``:

    >>> einsum('..., ...', 3, c)
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])
    >>> einsum(',ij', 3, c)
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])
    >>> einsum(3, [Ellipsis], c, [Ellipsis])
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])
    >>> 3 * c
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])

    Compute the inner product of ``b`` with itself, :math:`\sum_{i}{B_{i} B_{i}} = f`:

    >>> einsum('i,i', b, b)
    Tensor(30)
    >>> einsum(b, [0], b, [0])
    Tensor(30)
    >>> np.inner(b.data, b.data)
    30

    Compute the outer product of ``array([1, 2])`` with ``b``, :math:`A_{i}B_{j} = F_{ij}`:

    >>> einsum('i,j', np.arange(2)+1, b)
    Tensor([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> einsum(np.arange(2)+1, [0], b, [1])
    Tensor([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> np.outer(np.arange(2)+1, b)
    array([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> einsum('i...->...', a)
    Tensor([50, 55, 60, 65, 70])
    >>> einsum(a, [0,Ellipsis], [Ellipsis])
    Tensor([50, 55, 60, 65, 70])
    >>> np.sum(a, axis=0)
    array([50, 55, 60, 65, 70])

    Compute the tensor product :math:`\sum_{ij}{A_{ijk} B_{jil}} = F_{kl}`

    >>> a = mg.arange(60.).reshape(3,4,5)
    >>> b = mg.arange(24.).reshape(4,3,2)
    >>> einsum('ijk,jil->kl', a, b)
    Tensor([[ 4400.,  4730.],
            [ 4532.,  4874.],
            [ 4664.,  5018.],
            [ 4796.,  5162.],
            [ 4928.,  5306.]])
    >>> einsum(a, [0,1,2], b, [1,0,3], [2,3])
    Tensor([[ 4400.,  4730.],
            [ 4532.,  4874.],
            [ 4664.,  5018.],
            [ 4796.,  5162.],
            [ 4928.,  5306.]])
    >>> np.tensordot(a,b, axes=([1,0],[0,1]))
    array([[ 4400.,  4730.],
            [ 4532.,  4874.],
            [ 4664.,  5018.],
            [ 4796.,  5162.],
            [ 4928.,  5306.]])

    Matrix multiply ``a.T`` with ``b.T``, :math:`\sum_{k}{A_{ki} B_{jk}} = F_{ij}`

    >>> a = mg.arange(6).reshape((3,2))
    >>> b = mg.arange(12).reshape((4,3))
    >>> einsum('ki,jk->ij', a, b)
    Tensor([[10, 28, 46, 64],
            [13, 40, 67, 94]])
    >>> einsum('ki,...k->i...', a, b)
    Tensor([[10, 28, 46, 64],
            [13, 40, 67, 94]])
    >>> einsum('k...,jk', a, b)
    Tensor([[10, 28, 46, 64],
            [13, 40, 67, 94]])

    Make an assignment to a view along the diagonal of ``a``:

    >>> a = mg.zeros((3, 3))
    >>> einsum('ii->i', a).data[:] = 1
    >>> a
    Tensor([[ 1.,  0.,  0.],
            [ 0.,  1.,  0.],
            [ 0.,  0.,  1.]])
    """

    # TODO: normalize error handling for invalid inputs
    operands = list(operands)
    if isinstance(operands[0], str):
        # operands form: "ijk, ijk", x, y
        variables = operands[1:]
        if any(isinstance(i, Tensor) for i in operands):
            operands[1:] = (
                var.data if isinstance(var, Tensor) else var for var in operands[1:]
            )
    else:
        # operands form: op0, sublist0, op1, sublist1, ..., [sublistout]
        end = -1 if len(operands) % 2 else None  # -1 if sublistout is included
        variables = operands[:end:2]
        if any(isinstance(i, Tensor) for i in operands):
            operands[:end:2] = (
                var.data if isinstance(var, Tensor) else var for var in operands[:end:2]
            )

    in_lbls, out_lbls, _ = _parse_einsum_input(operands)

    # einsum doesn't handle out=None properly in numpy 1.17

    return Tensor._op(
        EinSum,
        *variables,
        op_kwargs=dict(in_lbls=in_lbls, out_lbls=out_lbls, optimize=optimize),
        constant=constant,
        out=out,
    )



from functools import reduce
from typing import TYPE_CHECKING

import numpy as np

from mygrad.operation_base import BinaryUfunc, Operation, UnaryUfunc

if TYPE_CHECKING:  # pragma: no cover
    from mygrad import Tensor

__all__ = [
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Reciprocal",
    "Power",
    "Square",
    "Positive",
    "Negative",
    "AddSequence",
    "MultiplySequence",
]

# Binary Ops


class Add(BinaryUfunc):
    numpy_ufunc = np.add

    def backward_var(self, grad, index, **kwargs):
        return grad


class Subtract(BinaryUfunc):
    numpy_ufunc = np.subtract

    def backward_var(self, grad, index, **kwargs):
        if index == 0:
            return grad
        else:
            return -grad


class Multiply(BinaryUfunc):
    numpy_ufunc = np.multiply

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:  # backprop through a
            return grad * b.data
        elif index == 1:  # backprop through b
            return grad * a.data


class Divide(BinaryUfunc):
    numpy_ufunc = np.divide

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:  # backprop through a
            return grad / b.data
        else:  # broadcast through b
            return -grad * a.data / (b.data ** 2)


class Power(BinaryUfunc):
    numpy_ufunc = np.power

    def backward_var(self, grad, index, **kwargs):
        x, y = (i.data for i in self.variables)

        if index == 0:
            return grad * y * (x ** np.where(y, (y - 1), 1))
        else:
            return grad * (x ** y) * np.log(np.where(x, x, 1))


# Unary Ops


class Reciprocal(UnaryUfunc):
    numpy_ufunc = np.reciprocal

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return -grad * np.reciprocal(a.data ** 2)


class Square(UnaryUfunc):
    numpy_ufunc = np.square

    def backward_var(self, grad, index, **kwargs):
        grad = 2 * grad
        grad *= self.variables[index].data
        return grad


class Positive(UnaryUfunc):
    numpy_ufunc = np.positive

    def backward_var(self, grad, index, **kwargs):
        return grad


class Negative(UnaryUfunc):
    numpy_ufunc = np.negative

    def backward_var(self, grad, index, **kwargs):
        return np.negative(grad)


class AddSequence(Operation):
    """Performs f(a, b, ..., z) = a + b + ... + z"""

    def __call__(self, *input_vars: "Tensor") -> np.ndarray:
        assert len(input_vars) > 1, "`add_sequence` requires at least two operands"
        self.variables = input_vars
        out = sum(var.data for var in input_vars)
        return out

    def backward_var(self, grad, index, **kwargs):
        return grad


class MultiplySequence(Operation):
    """ Performs f(a, b, ..., z) = a * b * ... * z"""

    def __call__(self, *input_vars: "Tensor") -> np.ndarray:
        self.variables = input_vars
        assert 2 <= len(self.variables)

        out = reduce(lambda x, y: x * y, (var.data for var in input_vars))
        self._iszero = np.any(out == 0)
        return out

    def backward(self, grad, *, graph, **kwargs):
        """Back-propagates the gradient through all of the operation's inputs. This needs to be updated
        by an operation if that operation takes more than 2 Tensor arguments."""
        if not self._iszero:
            self._product = grad * reduce(
                lambda x, y: x * y, (var.data for n, var in enumerate(self.variables))
            )
        else:
            self._product = None
        super().backward(grad, graph=graph, **kwargs)

    def backward_var(self, grad, index, **kwargs):
        var = self.variables[index]
        if not self._iszero:
            return self._product / var.data
        else:
            return grad * reduce(
                lambda x, y: x * y,
                (var.data for n, var in enumerate(self.variables) if n != index),
            )



import numpy as np

from mygrad import Tensor


def glorot_normal(*shape, gain=1, dtype=np.float32, constant=None):
    r"""Initialize a `Tensor` according to the normal initialization procedure
    described by Glorot and Bengio.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the output Tensor. Note that `shape` must be at least two-dimensional.

    gain : Real, optional (default=1)
        The gain (scaling factor) to apply.

    dtype : data-type, optional (default=float32)
        The data type of the output tensor; must be a floating-point type.

    constant : bool, optional (default=False)
        If `True`, the returned tensor is a constant (it
            does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values initialized according to the glorot normal initialization.

    Notes
    -----
    Glorot and Bengio put forward this initialization in the paper
        "Understanding the Difficulty of Training Deep Feedforward Neural Networks"
    http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf

    A Tensor :math:`W` initialized in this way should be drawn from a distribution about

    .. math::
        \mathcal{N}(0, \frac{\sqrt{2}}{\sqrt{n_j+n_{j+1}}})
    """
    if not np.issubdtype(dtype, np.floating):
        raise ValueError("Glorot Normal initialization requires a floating-point dtype")

    if len(shape) == 1:
        shape = shape[0]
    if len(shape) < 2:
        raise ValueError(
            "Glorot Normal initialization requires at least two dimensions"
        )

    if isinstance(gain, Tensor):
        gain = gain.item()

    fan_in = shape[1] * (shape[-1] if len(shape) > 2 else 1)
    fan_out = shape[0] * (shape[-1] if len(shape) > 2 else 1)
    std = gain * np.sqrt(2 / (fan_in + fan_out))
    return Tensor(
        np.random.normal(0, std, shape),
        dtype=dtype,
        constant=constant,
        copy=False,
    )



from typing import Optional, Union

from numpy import ndarray

from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike, DTypeLikeReals, Mask
from mygrad.ufuncs import ufunc_creator

from .ops import Exp, Exp2, Expm1, Log, Log1p, Log2, Log10, Logaddexp, Logaddexp2

__all__ = [
    "exp",
    "exp2",
    "expm1",
    "logaddexp",
    "logaddexp2",
    "log",
    "log2",
    "log10",
    "log1p",
]


@ufunc_creator(Exp)
def exp(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    r"""Calculate the exponential of all elements in the input tensor.

    This docstring was adapted from that of numpy.exp [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input values.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    exp : Tensor
        ``f(x1)`` computed element-wise

    See Also
    --------
    expm1 : Calculate ``exp(x) - 1`` for all elements in the tensor.
    exp2  : Calculate ``2**x`` for all elements in the tensor.

    Notes
    -----
    The irrational number ``e`` is also known as Euler's number.  It is
    approximately 2.718281, and is the base of the natural logarithm,
    ``ln`` (this means that, if :math:`x = \ln y = \log_e y`,
    then :math:`e^x = y`. For real input, ``exp(x)`` is always positive.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.exp.html

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.tensor(1.)
    >>> f = mg.exp(x); f  # f(1.)
    Tensor(2.71828183)

    Evaluate df/dx at ``x = 1``.

    >>> f.backward()
    >>> x.grad
    >>> x.grad  # df/dx @ x=1
    array(2.71828183)
    """
    ...


@ufunc_creator(Exp2)
def exp2(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Calculate `2**p` for all `p` in the input tensor.

    This docstring was adapted from that of numpy.exp2 [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input values.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    exp2 : Tensor
        ``f(x1)`` computed element-wise

    See Also
    --------
    power


    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.exp2.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.exp2([2., 3.])
    Tensor([ 4.,  8.])
    """
    ...


@ufunc_creator(Expm1)
def expm1(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Calculate ``exp(x) - 1`` for all elements in the tensor.

    This docstring was adapted from that of numpy.expm1 [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input values.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    expm1 : Tensor
        ``f(x1)`` computed element-wise

    See Also
    --------
    log1p : ``log(1 + x)``, the inverse of expm1.

    Notes
    -----
    This function provides greater precision than ``exp(x) - 1``
    for small values of ``x``.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.expm1.html

    Examples
    --------
    The true value of ``exp(1e-10) - 1`` is ``1.00000000005e-10`` to
    about 32 significant digits. This example shows the superiority of
    expm1 in this case.

    >>> import mygrad as mg
    >>> mg.expm1(1e-10)
    Tensor(1.00000000005e-10)
    >>> mg.exp(1e-10) - 1
    Tensor(1.000000082740371e-10)
    """
    ...


@ufunc_creator(Log)
def log(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Natural logarithm, element-wise.

    The natural logarithm ``log`` is the inverse of the exponential function,
    so that ``log(exp(x)) = x``. The natural logarithm is logarithm in base
    ``e``.

    This docstring was adapted from that of numpy.log [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input value.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    log : Tensor
        ``f(x1)`` computed element-wise

    See Also
    --------
    log10, log2, log1p

    Notes
    -----
    Logarithm is a multivalued function: for each `x` there is an infinite
    number of `z` such that `exp(z) = x`. The convention is to return the
    `z` whose imaginary part lies in `[-pi, pi]`.

    For real-valued input data types, `log` always returns real output. For
    each value that cannot be expressed as a real number or infinity, it
    yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `log` is a complex analytical function that
    has a branch cut `[-inf, 0]` and is continuous from above on it. `log`
    handles the floating-point negative zero as an infinitesimal negative
    number, conforming to the C99 standard.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.log.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.log([1, mg.e, mg.e**2, 0])
    array([  0.,   1.,   2., -Inf])
    """
    ...


@ufunc_creator(Log2)
def log2(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Base-2 logarithm applied elementwise to the tensor.

    This docstring was adapted from that of numpy.log2 [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input values.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    log2 : Tensor
        ``f(x1)`` computed element-wise

    See Also
    --------
    log, log10, log1p

    Notes
    -----
    For real-valued input data types, `log2` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.log2.html

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.tensor([0, 1, 2, 2**4])
    >>> mg.log2(x)
    Tensor([-Inf,   0.,   1.,   4.])
    """
    ...


@ufunc_creator(Log10)
def log10(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Return the base 10 logarithm of the input tensor, element-wise.

    This docstring was adapted from that of numpy.log10 [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input values.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    log10 : Tensor
        ``f(x1)`` computed element-wise

    Notes
    -----
    For real-valued input data types, `log10` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.log10.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.log10([1e-15, -3.])
    Tensor([-15.,  nan])
    """
    ...


@ufunc_creator(Log1p)
def log1p(
    x1: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Return the natural logarithm of one plus the input tensor, element-wise.

    Calculates ``log(1 + x)``.

    This docstring was adapted from that of numpy.log1p [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Input values.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    log1p : Tensor
        ``f(x1)`` computed element-wise

    See Also
    --------
    expm1 : ``exp(x) - 1``, the inverse of `log1p`.

    Notes
    -----
    For real-valued input, `log1p` is accurate also for `x` so small
    that `1 + x == 1` in floating-point accuracy.

    For real-valued input data types, `log1p` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.log1p.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.log1p(1e-99)
    1e-99
    >>> mg.log(1 + 1e-99)
    0.0
    """
    ...


@ufunc_creator(Logaddexp)
def logaddexp(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Logarithm of the sum of exponentiations of the inputs.

    Calculates ``log(exp(x1) + exp(x2))``. This function is useful in
    statistics where the calculated probabilities of events may be so small
    as to exceed the range of normal floating point numbers.  In such cases
    the logarithm of the calculated probability is stored. This function
    allows adding probabilities stored in such a fashion.

    This docstring was adapted from that of numpy.logaddexp [1]_

    Parameters
    ----------
    x1, x2 : ArrayLike
        Input values.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output).

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    logaddexp : Tensor
        Logarithm of ``exp(x1) + exp(x2)``.

    See Also
    --------
    logaddexp2: Logarithm of the sum of exponentiations of inputs in base 2.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.logaddexp.html

    Examples
    --------
    >>> import mygrad as mg
    >>> prob1 = mg.log(1e-50)
    >>> prob2 = mg.log(2.5e-50)
    >>> prob12 = mg.logaddexp(prob1, prob2)
    >>> prob12
    Tensor(-113.87649168120691)
    >>> mg.exp(prob12)
    Tensor(3.5000000000000057e-50)
    """
    ...


@ufunc_creator(Logaddexp2)
def logaddexp2(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Logarithm of the sum of exponentiations of the inputs in base-2.

    Calculates ``log2(2**x1 + 2**x2)``. This function is useful in machine
    learning when the calculated probabilities of events may be so small as
    to exceed the range of normal floating point numbers.  In such cases
    the base-2 logarithm of the calculated probability can be used instead.
    This function allows adding probabilities stored in such a fashion.

    This docstring was adapted from that of numpy.logaddexp2 [1]_

    Parameters
    ----------
    x1, x2 : ArrayLike
        Input values.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output).

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    logaddexp2 : Tensor
        Base-2 logarithm of ``2**x1 + 2**x2``.

    See Also
    --------
    logaddexp: Logarithm of the sum of exponentiations of the inputs.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.logaddexp2.html

    Examples
    --------
    >>> import mygrad as mg
    >>> prob1 = mg.log2(1e-50)
    >>> prob2 = mg.log2(2.5e-50)
    >>> prob12 = mg.logaddexp2(prob1, prob2)
    >>> prob1, prob2, prob12
    (Tensor(-166.09640474436813), Tensor(-164.77447664948076), Tensor(-164.28904982231052))
    >>> 2 ** prob12
    Tensor(3.4999999999999914e-50)
    """
    ...



import numpy as np

from mygrad import Tensor


def normal(*shape, mean=0, std=1, dtype=np.float32, constant=None):
    """Initialize a :class:`mygrad.Tensor` by drawing from a normal (Gaussian) distribution.

    Parameters
    ----------
    shape : Sequence[int]
        The output shape.

    mean : Real, optional (default=0)
        The mean of the distribution from which to draw.

    std : Real, optional (default=1)
        The standard deviation of the distribution from which to draw.

    dtype : data-type, optional (default=float32)
        The data type of the output tensor; must be a floating-point type.

    constant : bool, optional (default=False)
        If `True`, the returned tensor is a constant (it
            does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values drawn from Ɲ(μ, σ²), where μ=`mean` and σ=`std`.

    Examples
    --------
    >>> from mygrad.nnet.initializers import normal
    >>> normal(1, 2, 3)
    Tensor([[[-0.06481607, -0.550582  ,  0.04689528],
             [ 0.82973075,  2.83742   ,  1.0964519 ]]], dtype=float32)

    >>> normal(2, 2, dtype="float16", constant=True)
    Tensor([[-1.335 ,  0.9297],
            [ 1.746 , -0.1222]], dtype=float16)

    >>> normal(5, dtype="float64")
    Tensor([-0.03875407,  0.65368466, -0.72636993,  1.57404148, -1.17444345])
    """
    if not np.issubdtype(dtype, np.floating):
        raise ValueError("Normal initialization requires a floating-point dtype")
    if std < 0:
        raise ValueError("Standard deviation must be non-negative")

    if len(shape) == 1:
        shape = shape[0]

    if isinstance(mean, Tensor):
        mean = mean.item()
    if isinstance(std, Tensor):
        std = std.item()

    return Tensor(
        np.random.normal(mean, std, shape),
        dtype=dtype,
        constant=constant,
        copy=False,
    )



import numpy as np

from mygrad.operation_base import Operation, UnaryUfunc

__all__ = [
    "Sinh",
    "Cosh",
    "Tanh",
    "Csch",
    "Sech",
    "Coth",
    "" "Arcsinh",
    "Arccosh",
    "Arctanh",
    "Arccsch",
    "Arccoth",
]


class Sinh(UnaryUfunc):
    numpy_ufunc = np.sinh

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad * np.cosh(a.data)


class Cosh(UnaryUfunc):
    numpy_ufunc = np.cosh

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad * np.sinh(a.data)


class Tanh(UnaryUfunc):
    numpy_ufunc = np.tanh

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad * (1 - np.tanh(a.data) ** 2)


class Csch(Operation):
    """ f(a) -> csch(a) """

    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.sinh(a.data)

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad * -np.cosh(a.data) / np.sinh(a.data) ** 2


class Sech(Operation):
    """ f(a) -> sech(a) """

    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.cosh(a.data)

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad * -np.sinh(a.data) / np.cosh(a.data) ** 2


class Coth(Operation):
    """ f(a) -> coth(a) """

    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.tanh(a.data)

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad * -1 / np.sinh(a.data) ** 2


class Arcsinh(UnaryUfunc):
    numpy_ufunc = np.arcsinh

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad / np.sqrt(1 + a.data ** 2)


class Arccosh(UnaryUfunc):
    numpy_ufunc = np.arccosh

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad / np.sqrt(a.data ** 2 - 1)


class Arctanh(UnaryUfunc):
    numpy_ufunc = np.arctanh

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad / (1 - a.data ** 2)


class Arccsch(Operation):
    """ f(a) -> arccsch(a) """

    def __call__(self, a):
        self.variables = (a,)
        return np.arcsinh(1 / a.data)

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return -grad / (np.abs(a.data) * np.sqrt(1 + a.data ** 2))


class Arccoth(Operation):
    """ f(a) -> arccoth(a) """

    def __call__(self, a):
        self.variables = (a,)
        return np.arctanh(1 / a.data)

    def backward_var(self, grad, index, **kwargs):
        (a,) = self.variables
        return grad / (1 - a.data ** 2)



import numpy

__all__ = [
    "bool8",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "intp",
    "uintp",
    "float16",
    "float32",
    "float64",
    "half",
    "single",
    "double",
    "longdouble",
]

bool8 = numpy.bool8
int8 = numpy.int8
int16 = numpy.int16
int32 = numpy.int32
int64 = numpy.int64
uint8 = numpy.uint8
uint16 = numpy.uint16
uint32 = numpy.uint32
uint64 = numpy.uint64
intp = numpy.intp
uintp = numpy.uintp
float16 = numpy.float16
float32 = numpy.float32
float64 = numpy.float64
half = numpy.half
single = numpy.single
double = numpy.double
longdouble = numpy.longdouble



from typing import Optional, Sequence, Union

from numpy import ndarray

from mygrad.tensor_base import Tensor, implements_numpy_override
from mygrad.typing import ArrayLike, DTypeLikeReals

from .ops import Concatenate, Stack

__all__ = ["concatenate", "stack"]


@implements_numpy_override()
def concatenate(
    tensors: Sequence[ArrayLike],
    axis: Optional[int] = 0,
    out: Optional[Union[ndarray, Tensor]] = None,
    dtype: Optional[DTypeLikeReals] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    concatenate((t1, t2, ...), axis=0, out=None, *, constant=None)

    Join a sequence of tensors along an existing axis.

    This docstring was adapted from that of numpy.concatenate [1]_

    Parameters
    ----------
    tensors : Sequence[ArrayLike]
        The tensors must have the same shape, except in the dimension
        corresponding to `axis` (the first, by default).

    axis : Optional[int]
        The axis along which the tensors will be joined.  If axis is ``None``,
        tensors are flattened before use.  Default is 0.

    out : Optional[Union[ndarray, Tensor]]
        If provided, the destination to place the result. The shape must be
        correct, matching that of what concatenate would have returned if no
        out argument were specified.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        If provided, the destination array will have this dtype. Cannot be provided
        together with ``out``.

        Requires numpy 1.20 or higher.

    Returns
    -------
    res : Tensor
        The concatenated tensor.

    See Also
    --------
    stack : Stack a sequence of tensors along a new axis.
    hstack : Stack tensors in sequence horizontally (column wise).
    vstack : Stack tensors in sequence vertically (row wise).
    dstack : Stack tensors in sequence depth wise (along third dimension).

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.concatenate.html

    Examples
    --------
    >>> import mygrad as mg
    >>> a = mg.tensor([[1, 2], [3, 4]])
    >>> b = mg.tensor([[5, 6]])
    >>> mg.concatenate((a, b), axis=0)
    Tensor([[1, 2],
           [3, 4],
           [5, 6]])
    >>> mg.concatenate((a, b.T), axis=1)
    Tensor([[1, 2, 5],
           [3, 4, 6]])
    >>> mg.concatenate((a, b), axis=None)
    Tensor([1, 2, 3, 4, 5, 6])
    """
    return Tensor._op(
        Concatenate,
        *tensors,
        op_kwargs={"axis": axis, "dtype": dtype},
        constant=constant,
        out=out,
    )


@implements_numpy_override()
def stack(
    tensors: Sequence[ArrayLike],
    axis: int = 0,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    stack((t1, t2, ...), axis=0, out=None, *, constant=None)

    Join a sequence of tensors along a new axis.

    This docstring was adapted from that of numpy.stack [1]_

    Parameters
    ----------
    tensors : Sequence[ArrayLike]
        Each tensor must have the same shape.

    axis : Optional[int]
        The axis in the result tensor along which the input tensors are stacked.

    out : Optional[Union[ndarray, Tensor]]
        If provided, the destination to place the result. The shape must be
        correct, matching that of what concatenate would have returned if no
        out argument were specified.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    res : Tensor
        The stacked tensor has one more dimension than the input arrays.

    See Also
    --------
    concatenate : Join a sequence of tensors along an existing axis.
    hstack : Stack tensors in sequence horizontally (column wise).
    vstack : Stack tensors in sequence vertically (row wise).
    dstack : Stack tensors in sequence depth wise (along third dimension).

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.stack.html

    Examples
    --------
    >>> import mygrad as mg
    >>> a = mg.tensor([1, 2, 3])
    >>> b = mg.tensor([-1, -2, -3])
    >>> mg.stack((a, b))
    Tensor([[ 1,  2,  3],
            [-1, -2, -3]])

    >>> mg.stack((a, b), axis=-1)
    Tensor([[1, -1],
            [2, -2],
            [3, -3]])
    """
    return Tensor._op(
        Stack,
        *tensors,
        op_kwargs={"axis": axis},
        constant=constant,
        out=out,
    )



from numbers import Real
from typing import Optional

import numpy as np

import mygrad._utils.graph_tracking as _tracking
from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor, asarray
from mygrad.typing import ArrayLike


class MarginRanking(Operation):
    def __call__(self, x1, x2, y, margin):
        """Computes the margin ranking loss between ``x1``
        and ``x2``.

        Parameters
        ----------
        x1 : mygrad.Tensor, shape=(N,) or (N, D)

        x2 : mygrad.Tensor, shape=(N,) or (N, D)

        y : numpy.ndarray

        margin : float

        Returns
        -------
        numpy.ndarray, shape=()
        """
        self.variables = (x1, x2)
        x1 = x1.data
        x2 = x2.data

        self.y = y

        M = margin - self.y * (x1 - x2)
        not_thresh = M <= 0
        loss = M
        loss[not_thresh] = 0.0
        if _tracking.TRACK_GRAPH:
            self._grad = np.ones_like(M)
            self._grad[not_thresh] = 0.0
            self._grad /= M.size
        return np.mean(loss)

    def backward_var(self, grad, index, **kwargs):
        sign = -self.y if index == 0 else self.y
        return grad * (sign * self._grad)


def margin_ranking_loss(
    x1: ArrayLike,
    x2: ArrayLike,
    y: ArrayLike,
    margin: float,
    *,
    constant: Optional[bool] = None
) -> Tensor:
    r"""Computes the margin average margin ranking loss.
    Equivalent to::

    >>> import mygrad as mg
    >>> mg.mean(mg.maximum(0, margin - y * (x1 - x2)))

    Parameters
    ----------
    x1 : ArrayLike, shape=(N,) or (N, D)
        A batch of scores or descriptors to compare against those in `x2`

    x2 : ArrayLike, shape=(N,) or (N, D)
        A batch of scores or descriptors to compare against those in `x1`

    y  : Union[int, ArrayLike], scalar or shape=(N,)
        1 or -1. Specifies whether the margin is compared against `(x1 - x2)`
        or `(x2 - x1)`, for each of the N comparisons.

    margin : float
        A non-negative value to be used as the margin for the loss.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor, shape=()
        The mean margin ranking loss.
    """
    if not 0 < x1.ndim < 3:
        raise ValueError("`x1` must have shape (N,) or (N, D)")
    if not x1.shape == x2.shape:
        raise ValueError("`x1` and `x2` must have the same shape")
    if not np.issubdtype(x1.dtype, np.floating):
        raise TypeError("`x1` must contain floats")
    if not np.issubdtype(x2.dtype, np.floating):
        raise TypeError("`x2` must contain floats")
    if not isinstance(margin, Real) or margin < 0:
        raise ValueError("`margin` must be a non-negative scalar")

    y = asarray(y)

    if y.size == 1:
        y = np.array(y.item())

    if not y.ndim == 0 and not (y.ndim == 1 and len(y) == len(x1)):
        raise ValueError("`y` must be a scalar or shape-(N,) array of ones")

    if y.ndim:
        if x1.ndim == 2:
            y = y.reshape(-1, 1)
    return Tensor._op(MarginRanking, x1, x2, op_args=(y, margin), constant=constant)



from typing import Optional

from numpy import ndarray

from mygrad import Tensor, multiply
from mygrad.typing import ArrayLike

from .sigmoid import sigmoid


def glu(x: ArrayLike, axis: int = -1, *, constant: Optional[bool] = None) -> Tensor:
    """Returns the Gated Linear Unit A * σ(B), where A and B are split from `x`.

    Parameters
    ----------
    x : ArrayLike
        The input.

    axis : int, optional (default=-1)
        The axis along which to split the input in half and apply the GLU.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor
        The result of applying the  Gated Linear Unit elementwise to the input.

    Notes
    -----
    The Gated Linear Unit was proposed in the paper
        "Language Modeling with Gated Convolutional Networks"
        Yann Dauphin, Angela Fan, Michael Auli, David Grangier
    available at https://arxiv.org/abs/1612.08083

    The GLU operation splits the input `x` in half along `axis`, storing the first half in A and the
    second in B. The return value is then A ⊙ σ(B), where ⊙ is elementwise multiplication and σ is
    the sigmoid function.

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet.activations import glu
    >>> x = mg.arange(-5., 5.)
    >>> x
    Tensor([-5., -4., -3., -2., -1.,  0.,  1.,  2.,  3.,  4.])
    >>> y = glu(x); y
    Tensor([-2.5       , -2.92423431, -2.64239123, -1.90514825, -0.98201379])
    >>> y.backward()
    >>> x.grad
    array([ 0,  0,  0,  0,  0, -1,  0,  0,  0,  0])
    """
    if isinstance(axis, (ndarray, Tensor)):
        axis = axis.item()

    if not isinstance(axis, int):
        raise TypeError(
            f"`axis` must be an integer-valued scalar, got {axis} (type {type(axis)})"
        )

    first_idx = list(slice(None) for _ in x.shape)
    second_idx = list(slice(None) for _ in x.shape)
    first_idx[axis] = slice(0, x.shape[axis] // 2)
    second_idx[axis] = slice(x.shape[axis] // 2, None)

    first_half = x[tuple(first_idx)]
    second_half = x[tuple(second_idx)]

    if first_half.shape != second_half.shape:
        raise ValueError(
            f"The shapes after splitting must be the same but got {first_half.shape} "
            "and {second_half.shape}"
        )
    return multiply(first_half, sigmoid(second_half), constant=constant)



from typing import Optional

import numpy as np

from mygrad import Tensor
from mygrad.operation_base import Operation
from mygrad.typing import ArrayLike, Real

__all__ = ["selu"]


_ALPHA = 1.6732632423543772848170429916717
_SCALE = 1.0507009873554804934193349852946


class SELU(Operation):
    """Returns the scaled exponential linear activation (SELU) elementwise along x. The SELU is
    given by  λɑ(exp(x) - 1) for x < 0 and λx for x ≥ 0.

    Notes
    -----
    The SELU activation was proposed in the paper
        Self-Normalizing Neural Networks
        Günter Klambauer, Thomas Unterthiner, Andreas Mayr, Sepp Hochreiter
    at https://arxiv.org/abs/1706.02515
    """

    def __call__(self, x):
        """
        Parameters
        ----------
        x : mygrad.Tensor
            Input data.

        Returns
        -------
        numpy.ndarray
            The SELU function applied to `x` elementwise.
        """
        self.variables = (x,)

        x = x.data
        self.exp = _ALPHA * (np.exp(x) - 1)
        return _SCALE * np.where(x < 0, self.exp, x)

    def backward_var(self, grad, index, **kwargs):
        x = self.variables[index]
        return grad * _SCALE * np.where(x.data < 0, self.exp + _ALPHA, 1)


def selu(x: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """Returns the scaled exponential linear activation (SELU) elementwise along x.

    The SELU is given by  λɑ(exp(x) - 1) for x < 0 and λx for x ≥ 0.

    Parameters
    ----------
    x : ArrayLike
        Input data.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor
        The SELU function applied to `x` elementwise.

    References
    ----------
    .. [1] Günter Klambauer, Thomas Unterthiner, Andreas Mayr, Sepp Hochreiter
       Self-Normalizing Neural Networks
       https://arxiv.org/abs/1706.02515

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet.activations import selu
    >>> x = mg.arange(-5, 6)
    >>> x
    Tensor([-5, -4, -3, -2, -1,  0,  1,  2,  3,  4,  5])
    >>> y = selu(x, alpha=0.1); y
    Tensor([-1.74625336, -1.72589863, -1.67056873, -1.52016647, -1.11133074,
         0.        ,  1.05070099,  2.10140197,  3.15210296,  4.20280395,
         5.25350494])

    .. plot::

       >>> import mygrad as mg
       >>> from mygrad.nnet.activations import selu
       >>> import matplotlib.pyplot as plt
       >>> x = mg.linspace(-2, 2, 100)
       >>> y = selu(x)
       >>> plt.title("selu(x)")
       >>> y.backward()
       >>> plt.plot(x, x.grad, label="df/dx")
       >>> plt.plot(x, y, label="f(x)")
       >>> plt.legend()
       >>> plt.grid()
       >>> plt.show()
    """
    return Tensor._op(SELU, x, constant=constant)



from typing import Optional

from mygrad import abs, divide
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike

__all__ = ["soft_sign"]


def soft_sign(x: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """Returns the soft sign function x / (1 + |x|).

    Parameters
    ----------
    x : ArrayLike
        Input data.

    constant : boolean, optional (default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor
        The soft sign function applied to `x` elementwise.

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet.activations import soft_sign
    >>> x = mg.arange(-5, 6)
    >>> x
    Tensor([-5, -4, -3, -2, -1,  0,  1,  2,  3,  4,  5])
    >>> y = soft_sign(x); y
    Tensor([-0.83333333, -0.8       , -0.75      , -0.66666667, -0.5       ,
         0.        ,  0.5       ,  0.66666667,  0.75      ,  0.8       ,
         0.83333333])

    .. plot::

       >>> import mygrad as mg
       >>> from mygrad.nnet.activations import soft_sign
       >>> import matplotlib.pyplot as plt
       >>> x = mg.linspace(-10, 10, 100)
       >>> y = soft_sign(x)
       >>> plt.title("soft_sign(x)")
       >>> y.backward()
       >>> plt.plot(x, x.grad, label="df/dx")
       >>> plt.plot(x, y, label="f(x)")
       >>> plt.legend()
       >>> plt.grid()
       >>> plt.show()
    """
    return divide(x, 1 + abs(x), constant=constant)
